/*********************************************************************/
/* File: environment.cc - functions for the environment handling     */
/*                            HUT TINTIN                             */
/*                     coded by Ossi Tulonen                         */
/*********************************************************************/
/*
 * Copyright (c) 1996 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define __ENVIRONMENT_C__ 1

#include "structs.h"

#include <string.h>
#include <pwd.h>
#include <assert.h>

#include "session.h"

#include "environment.h"
#include "net.h"
#include "utility.h"

#if !defined(environ)
extern char **environ;
#endif

extern const uchar *version_num;

env_list environment_list;

env_elem::env_elem(const uchar *env_name, const uchar *env_value)
{
  name  = (uchar *)strdup((const char *)env_name);
  value = (uchar *)strdup((const char *)env_value);
  
  well_known = FALSE;
  
  next = 0;
  prev = 0;
}

env_elem::~env_elem()
{
  if(name)
    free(name);
  
  if(value)
    free(value);
}

env_list::env_list()
{
  char **env, *str;
  env_elem *ee;
  
  res_line = 0;
  len = 0;
  size = 0;
  
  envs = 0;
  
  for(env = environ; *env; env++) {
    
    if((str = strchr(*env, '='))) {
      *str = '\0';
      
      add((const uchar *)*env, (const uchar *)str + 1);
      *str = '=';
    }
  }
  
  if((ee = find((const uchar *)"DISPLAY")) && *ee->value == ':') {
    uchar hostname[1024];
    
    gethostname((char *)hostname, sizeof(hostname) - 10);
    str_cat(hostname, strlen((const char *)hostname), sizeof(hostname),
	    ee->value);
    add((const uchar *)"DISPLAY", hostname);
  }

  if(!find((const uchar *)"USER")) {
    
    if((ee = find((const uchar *)"LOGNAME"))) {
      add((const uchar *)"USER", ee->value);
      
    } else {
      struct passwd *pw;
      
      if((pw = getpwuid(getuid())))
	add((const uchar *)"USER", (const uchar *)pw->pw_name);
    }
  }

  add((const uchar *)"HTVERSION", version_num);
  
  set_well_known((const uchar *)"USER");
  set_well_known((const uchar *)"JOB");
  set_well_known((const uchar *)"ACCT");
  set_well_known((const uchar *)"PRINTER");
  set_well_known((const uchar *)"SYSTEMTYPE");
  set_well_known((const uchar *)"DISPLAY");
}

env_list::~env_list()
{
  env_elem *ee, *next_ee;
  
  if(res_line)
    free(res_line);
  
  for(ee = envs; ee; ee = next_ee) {
    next_ee = ee->next;
    
    delete ee;
  }
}

void env_list::add(const uchar *name, const uchar *value)
{
  env_elem *ee;
  
  remove(name); // make sure it isn't in list twice 
  
  if(!(ee = new env_elem(name, value))) {
    filelog("out of memory");
    return;
  }
  
  ee->next = envs;
  envs = ee;
  if(envs->next)
    envs->next->prev = ee;
}

void env_list::remove(const uchar *name)
{
  env_elem *ee;
  
  if(!(ee = find(name)))
    return;
  
  // remove from list
  
  if(ee->prev) {
    ee->prev->next = ee->next;
    if(ee->next)
      ee->next->prev = ee->prev;
    
  } else {
    envs = ee->next;
    if(ee->next)
      ee->next->prev = 0;
  }
  
  delete ee;
}

void env_list::set_well_known(const uchar *name)
{
  env_elem *ee;
  
  if((ee = find(name)))
    ee->well_known = TRUE;
}

env_elem *env_list::find(const uchar *name) const
{
  env_elem *ee;
  
  for(ee = envs; ee; ee = ee->next) {
    if(!strcmp((const char *)ee->name, (const char *)name))
      return(ee);
  }
  
  return(0);
}

const uchar *env_list::find_value(const uchar *name) const
{
  env_elem *ee;
  
  if(!(ee = find(name)))
    return(0);
  
  return(ee->value);
}

void env_list::start_sb_is(int telopt, int &res_var_nr)
{
  res_var_nr = 0;
  
  assert(telopt == TELOPT_NEW_ENVIRON || telopt == TELOPT_OLD_ENVIRON);
  assert(!res_line && len == 0 && size == 0);
  
  len  = 0;
  size = 256;
  
  if(!(res_line = (uchar *)calloc(size, sizeof(uchar)))) {
    filelog("out of memory");
    
    size = 0;
    return;
  }
  
  if(telopt == TELOPT_NEW_ENVIRON) {
    var_nr   = NEW_ENV_VAR;
    value_nr = NEW_ENV_VALUE;
  } else {
    var_nr   = OLD_ENV_VAR;
    value_nr = OLD_ENV_VALUE;
  }
  res_var_nr = var_nr;
  
  res_line[len++] = IAC;
  res_line[len++] = SB;
  res_line[len++] = telopt;
  res_line[len++] = TELQUAL_IS;
}

void env_list::add_sb_is(const env_elem *e)
{
  uchar *str, *val_str;
  
  if(!e->name || !*e->name || !e->value || !*e->value || !res_line)
    return;
  
  if(e->is_well_known())
    res_line[len++] = var_nr;
  else
    res_line[len++] = ENV_USERVAR;
  
  str     = e->name;
  val_str = e->value;
  while(TRUE) {
    while(*str) {
      
      if(len + 6 > size) {
	size += 256;
	if(!(res_line = (uchar *)realloc(res_line, size))) {
	  filelog("out of memory");

	  len  = 0;
	  size = 0;
	  
	  return;
	}
      }
      
      if(*str == IAC) {
	res_line[len++] = IAC;
	
      } else if(*str == var_nr  || *str == value_nr ||
		*str == ENV_ESC || *str == ENV_USERVAR) {
	res_line[len++] = ENV_ESC;
      }
      
      res_line[len++] = *str++;
    }
    if((str = val_str)) {
      res_line[len++] = value_nr;
      val_str = 0;
      
    } else
      break;
  }
}

void env_list::add_all_sb_is(const int uservar)
{
  env_elem *ee;
  
  for(ee = envs; ee && res_line; ee = ee->next) {
    if((ee->well_known && !uservar) || (!ee->well_known && uservar))
      add_sb_is(ee);
  }
}

int env_list::end_sb_is(int emptyok, session *ses)
{
  if(!res_line)
    return(FALSE);
  
  if(emptyok || len > 6) {
    res_line[len++] = IAC;
    res_line[len++] = SE;
    res_line[len] = '\0';
    
    if(write_socket(res_line, ses->get_socket(), len) < 0) {
      ses->print_errmsg("write socket");
      ses->rem_net();
      len  = 0;
      size = 0;
      free(res_line);
      res_line = 0;
      return(-1);
    }
    len  = 0;
    size = 0;
    free(res_line);
    res_line = 0;
    
    return(TRUE);
  }
  
  len  = 0;
  size = 0;
  free(res_line);
  res_line = 0;
  
  return(FALSE);
}

const uchar *env_list::parse_name(const uchar *str, const uchar *str_end,
				  uchar *name) const
{
  while(str < str_end) {
    if(*str == var_nr || *str == value_nr || *str == ENV_USERVAR) {
      *name = '\0';
      return(str);
      
    } else if(*str == ENV_ESC || *str == IAC) {
      str++;
      if(str < str_end)
	*name++ = *str++;
      
    } else {
      *name++ = *str++;
    }
  }
  *name = '\0';
  
  return(str);
}
