/*********************************************************************/
/* file: net.c - do all the net stuff                                */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#define __HT_NET_C__ 1

#include "structs.h"

#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>

#include <arpa/inet.h>

#ifndef O_NONBLOCK
#  ifndef FNDELAY
#    define FNDELAY O_NDELAY
#  endif
#  define O_NONBLOCK FNDELAY
#endif

#include "types.h"

#include "action.h"
#include "environment.h"
#include "llist.h"
#include "main.h"
#include "net.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#include "telnet_opt.h"
#ifdef TERM
#  include "term.h"
#endif
#include "ticks.h"
#include "utility.h"
#include "utils.h"
#ifndef X_DISPLAY_MISSING
#include "xscreen.h"
#endif

#ifndef BADSIG
#  define BADSIG (void (*)(...))-1
#endif

extern int sessionsstarted, ticker_interrupted, window_interrupted;
extern struct master_table_data master_table;
extern int errno;
extern int signalmask;
extern int ownsignalmask;

#ifdef TERM
#  define CONNECT(sock, addr, len)       term_connect(sock, addr, len)
#  define CLOSE(sock)                    term_close(sock)
#  define GETHOSTBYADDR(addr, len, type) term_gethostbyaddr(addr, len, type)
#  define GETHOSTBYNAME(name)            term_gethostbyname(name)
#else
#  define CONNECT(sock, addr, len)       connect(sock, addr, len)
#  define CLOSE(sock)                    close(sock)
#  define GETHOSTBYADDR(addr, len, type) gethostbyaddr(addr, len, type)
#  define GETHOSTBYNAME(name)            gethostbyname(name)
#endif

// ****************************************
//  code to connect and disconnect session
// ****************************************

session *comm_connect(const uchar *arg, session *ses)
{
  uchar host[BUFFER_SIZE], port[BUFFER_SIZE];
  struct servent *sp;

  if(ses->get_socket() != -1) {
    ses->print((const uchar *)"#Already connected somewhere.\n\r");
    return(ses);
  }
  
  arg = get_arg_stop_spaces(arg, host);
  arg = get_arg_stop_spaces(arg, port);
  
  if(!*host) {
    ses->print((const uchar *)"#HEY! SPECIFY AN ADDRESS WILL YOU?\n\r");
    return(0);
  }
  
  if(!*port) {
    if(!(sp = getservbyname("telnet", "tcp"))) {
      ses->print((const uchar *)"# tcp/telnet: unknown service\n\r");
      return(0);
    }
    sprintf((char *)port, "%d", ntohs(sp->s_port));
    ses->print((const uchar *)"#Using default telnet port.\n\r");
  }
  
  connect_mud(host, port, ses);
  
  return(ses);
}

session *comm_disconnect(const uchar *, session *ses)
{
#ifndef X_DISPLAY_MISSING
  int sock;
#endif
  
  if(ses->get_socket() < 0) {
    ses->print((const uchar *)"#Not connected anywhere.\n\r");
    return(ses);
  }
#ifndef X_DISPLAY_MISSING
  if(ses->is_trycon()) {
    sock = ses->get_socket();
    ses->rem_ConNet();
    
    ses->set_addr(0, 0, 0);
    ses->print((const uchar *)"breaked.\n\r");
    CLOSE(sock);
    ses->clear_output_stack();
    ses->set_prompt((const uchar *)"> ");
  } else
#endif
    ses->rem_net();
  
  return(ses);
}

/**************************************************/
/* try connect to the mud specified by the args   */
/* return fd on success / 0 on failure            */
/**************************************************/
#define MAX_HOSTNAME   256

session *cur_con_try;

void break_connect(...)
{
  cur_con_try->print((const uchar *)"breaking...");
}

void send_errormsg(session *ses)
{
  switch(errno) {
  case EINTR:
    ses->print((const uchar *)"Breaked.\n\r");
    break;
  case ECONNREFUSED:
    ses->print((const uchar *)"CONNECTION REFUSED.\n\r");
    break;
  case ENETUNREACH:
    ses->print((const uchar *)"THE NETWORK IS NOT REACHABLE FROM THIS HOST.\n\r");
    break;
  case EINVAL:
    ses->print((const uchar *)"Unknown host.\n\r");
    break;
  default:
    ses->print((const uchar *)"Couldn't connect\n\r");
  }
}

#if !defined(SOL_IP) && defined(IPPROTO_IP)
#define SOL_IP IPPROTO_IP
#endif

void connect_mud(const uchar *hostname, unsigned char *port,
		 struct session *ses)
{
  const char *name = 0;
  unsigned char buffer[2048];
  int sock, connectresult, opt, res;

#if defined(SOL_IP) && defined(IP_TOS) && defined(IPTOS_LOWDELAY)
  int tos;
#endif
  
  unsigned long addr;
  struct sockaddr_in sockaddr;
  struct hostent *host;
#ifdef HAVE_SIGACTION
  struct sigaction new_siga, old_siga;
#endif
#ifdef X_DISPLAY_MISSING
  int timeout, null = 0;
  fd_set rs, es;
  struct timeval nulltime;
#endif

  if(isdigit(*port))  
    sockaddr.sin_port = htons(atoi((const char *)port));
  /* inteprete port part */
  else {
    ses->print((const uchar *)"#THE PORT SHOULD BE A NUMBER.\n\r");
    return;
  }
  
#ifndef X_DISPLAY_MISSING
  if(MainDisplay) {
    DrawFunc();
    XFlush(MainDisplay);       // flush buffers so user get return
  }
#endif
  
  if(isdigit(*hostname) &&
     (addr = inet_addr((const char *)hostname)) != (ulong)-1) {
    if(!(host = GETHOSTBYADDR((char *) &addr, sizeof(addr), AF_INET))) {
      sockaddr.sin_addr.s_addr = addr;
      name = "<Unknown>";
    } else {
      memcpy((char *)&sockaddr.sin_addr, host->h_addr,
	     sizeof(sockaddr.sin_addr));
      name = host->h_name;
    }
  } else if((host = GETHOSTBYNAME((char *)hostname))) {
    memcpy((char *)&sockaddr.sin_addr, host->h_addr,
	   sizeof(sockaddr.sin_addr));
    name = host->h_name;
  } else {
    sprintf((char *)buffer, "#ERROR - Unknown host %s\n\r", hostname);
    ses->print(buffer);
    return;
  }
  sockaddr.sin_family = AF_INET;
  
  if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    ses->print_errmsg("Couldn't get socket");
    return;
  }
  opt = 1;
  setsockopt(sock, SOL_SOCKET, SO_OOBINLINE, (char *)&opt, sizeof(opt));
  
#if defined(SOL_IP) && defined(IP_TOS) && defined(IPTOS_LOWDELAY)
  
  tos = IPTOS_LOWDELAY;
  setsockopt(sock, SOL_IP, IP_TOS, &tos, sizeof(tos));
  
#endif /* defined(SOL_IP) && defined(IP_TOS) */
  
  if((res = fcntl(sock, F_GETFL, 0)) < 0 ||
     fcntl(sock, F_SETFL, res | O_NONBLOCK) < 0) {
    ses->print_errmsg("SETTING NONBLOCK");
    CLOSE(sock);
    return;
  }

  ses->set_prompt((const uchar *)"#Trying to connect...");
  
  cur_con_try = ses;    // This must be set before calling break_conect()
  

#ifndef X_DISPLAY_MISSING
  if(MainDisplay) {
    DrawFunc();
    XFlush(MainDisplay);
  } else {
#endif
#ifdef HAVE_SIGACTION
    bzero((char *)&new_siga, sizeof(new_siga));
    new_siga.sa_flags   = SA_RESTART;
    
    new_siga.sa_handler = break_connect;
    if(sigaction(SIGINT, &new_siga, &old_siga))
      syserr("signal SIGINT");
#else
    if(signal(SIGINT, break_connect) == BADSIG)
      syserr("signal SIGINT");
#endif
    
    sigsetmask(ownsignalmask);
    nocbreak_echo();
#ifndef X_DISPLAY_MISSING
  }
#endif

  do {
    window_interrupted = FALSE;
    connectresult = CONNECT(sock,
			    (struct sockaddr *)&sockaddr,
			    sizeof(sockaddr));
  } while(connectresult < 0 && errno != EINPROGRESS && window_interrupted);
  
#ifndef X_DISPLAY_MISSING
  if(!MainDisplay) {
#endif
    cbreak_noecho();
    sigsetmask(0);
    
#ifdef HAVE_SIGACTION
    if(sigaction(SIGINT, &old_siga, 0))
      syserr("signal SIGINT");
#else
    if(signal(SIGINT, myquitsig) == BADSIG)
      syserr("signal SIGINT");
#endif
    
#ifndef X_DISPLAY_MISSING
  }
#endif
  
  cur_con_try = 0;
  
  if(connectresult < 0) {
    if(errno == EINPROGRESS) {
#ifndef X_DISPLAY_MISSING
      ses->add_prompt((const uchar *)"(use #disconnect command to break)...");
      
      ses->set_ConNet(sock);
      ses->set_addr((const uchar *)name,
		    (const uchar *)inet_ntoa(sockaddr.sin_addr),
		    ntohs(sockaddr.sin_port));
      return;
#else
      ses->print(ses->get_prompt());

      timeout = 180; // Wait 180 secs
      nulltime.tv_sec  = 0;
      nulltime.tv_usec = 0;
      
      ses->print((const uchar *)"(press any key to break)...");
      
      while(timeout > 0) {
	sleep(1);
	timeout--;
	
	FD_ZERO(&rs);
	FD_SET(0, &rs); /* stdin */
	
	sigsetmask(signalmask);
	res = select(1, &rs, (fd_set *)0, (fd_set *)0, &nulltime);
	sigsetmask(0);
	
	if(res > 0 && FD_ISSET(0, &rs)) {  // user has wrote something
	  read(0, buffer, sizeof(buffer)); // read input buffer empty
	  ses->print((const uchar *)"breaked.\n\r");
	  CLOSE(sock);
	  ses->set_prompt((const uchar *)"> ");
	  return;
	}
	
	FD_ZERO(&rs);
	FD_ZERO(&es);
	FD_SET(sock, &rs);
	FD_SET(sock, &es);
	
	sigsetmask(signalmask);
	res = select(sock + 1, &rs, (fd_set *)0, &es, &nulltime);
	sigsetmask(0);
	
	if(res > 0) {
	  if(FD_ISSET(sock, &rs)) {
	    if(recvfrom(sock, buffer, 1, MSG_PEEK, 0, &null) < 0) {
	      send_errormsg(ses);
	      CLOSE(sock);
	      ses->set_prompt((const uchar *)"> ");
	      return;
	    }
	  }
	  if(FD_ISSET(sock, &es)) {
	    send_errormsg(ses);
	    CLOSE(sock);
	    ses->set_prompt((const uchar *)"> ");
	    return;
	  }
	  break;
	}
      }
      if(timeout <= 0) {
	ses->print((const uchar *)"CONNECTION TIMED OUT.\n\r");
	CLOSE(sock);
	ses->set_prompt((const uchar *)"> ");
	return;
      }
#endif
    } else {
      ses->print(ses->get_prompt());
      send_errormsg(ses);
      CLOSE(sock);
      ses->set_prompt((const uchar *)"> ");
      return;
    }
  } else {
    ses->print(ses->get_prompt());
  }
  
  ses->print((const uchar *)"Ok\n\r");
  ses->set_prompt((const uchar *)"");
  
  sprintf((char *)buffer, "#Connected to %s (%s) port %d.\n\r",
	  inet_ntoa(sockaddr.sin_addr), name, ntohs(sockaddr.sin_port));
  ses->print(buffer);
  
  ses->set_net(sock);
  ses->set_addr((const uchar *)name,
		(const uchar *)inet_ntoa(sockaddr.sin_addr),
		ntohs(sockaddr.sin_port));
}

#ifndef X_DISPLAY_MISSING
void XConNet(XtPointer point, int *, XtInputId *)
{
  char buffer[2048];
  int null = 0, sock;
  session *ses = (session *)point;
  struct output_stack_elem *tmp;

  sock = ses->get_socket();
  
  ses->rem_ConNet();
  
  if(recvfrom(sock, buffer, 1, MSG_PEEK, 0, &null) < 0) {
    ses->set_addr(0, 0, 0);
    send_errormsg(ses);
    ses->set_prompt((const uchar *)"> ");
    CLOSE(sock);
    
    ses->clear_output_stack();
    
    ses->print((const uchar *)"> ");
    DrawFunc();
    return; /* couldn't connect */
  }
  
  ses->print((const uchar *)"Ok\n\r");
  ses->set_prompt((const uchar *)"");
  
  ses->set_net(sock);
  
  sprintf(buffer, "#Connected to %s (%s) %d.\n\r",
	  ses->get_host(), ses->get_ip_host(), ses->get_port());
  ses->print((const uchar *)buffer);
  
  for(tmp = ses->get_output_stack(); tmp && ses; tmp = tmp->next) {
    ses = ses->write_net(tmp->comm_line);
  }
  if(ses)
    ses->clear_output_stack();
  
  DrawFunc();
}
#endif

// ****************************************
//  code to print telnet unsupport message
// ****************************************

#ifdef TELCOM_DEBUG

void un_sup_tel_opt(int telcom, int telopt, session *ses)
{
  unsigned char buf[256];
  
#if defined(TELCMD_OK) && defined(TELCMD)
  sprintf((char *)buf,
	  "#UNSUPPORTED TELNET OPTION: com:%d(%s) opt:%d(%s)\n\r",
	  telcom,
	  TELCMD_OK(telcom) ? TELCMD(telcom) : "unknown",
	  telopt,
	  TELOPT_OK(telopt) ? TELOPT(telopt) : "unknown");
#else
  sprintf((char *)buf, "#UNSUPPORTED TELNET OPTION: com:%d opt:%d(%s)\n\r",
	  telcom,
	  telopt,
	  TELOPT_OK(telopt) ? TELOPT(telopt) : "unknown");
#endif
  ses->print(buf);
  ses->print((const uchar *)"#If you got this message with some dikumud email to Ossi.Tulonen@hut.fi\n\r");
  ses->print((const uchar *)"#those com and opt numbers. (please)\n\r");
}

void un_sup_tel_com(int telcom, session *ses)
{
  unsigned char buf[256];
  
#if defined(TELCMD_OK) && defined(TELCMD)
  sprintf((char *)buf,
	  "#UNSUPPORTED TELNET COMMAND: com:%d(%s)\n\r",
	  telcom,
	  TELCMD_OK(telcom) ? TELCMD(telcom) : "unknown");
#else
  sprintf((char *)buf, "#UNSUPPORTED TELNET COMMAND: com:%d\n\r",
	  telcom);
#endif
  ses->print(buf);
  ses->print((const uchar *)"#If you got this message with some dikumud email to Ossi.Tulonen@hut.fi\n\r");
  ses->print((const uchar *)"#that com number. (please)\n\r");
}

#endif

/************************************************************/
/* write line to the mud ses is connected to - add \n first */
/************************************************************/
int write_socket(const unsigned char *line, const int net, const int size)
{
  int did_put, sofar;
  
  assert(net >= 0);
  
  sofar = 0;
  do {    /* write until entire string has been written */

    /*    did_put = send(net, line + sofar, size - sofar, MSG_OOB); */
    
    did_put = write(net, line + sofar, size - sofar);
    
    if(did_put < 0)
      return(-1);
    
    sofar += did_put;
  } while( sofar < size);
  
  return(sofar);
}

session *session::write_net(const unsigned char *line)
{
  unsigned char outtext[BUFFER_SIZE + 1];
  int size;
  
  size = str_cat(outtext, 0, sizeof(outtext), line);
  size = str_cat(outtext, size, sizeof(outtext), (const uchar *)"\n");
  
  if(write_socket(outtext, net, size) < 0) {
    print_errmsg("write socket");
    rem_net();
    return(0);
  }
  
  if(master_table.ucommand) {
    set_prompt((const uchar *)"");
  }
  return(this);
}

session *session::send_naws()
{
  unsigned char tmp[20];
  int pos, lines;
  
  if(net >= 0 && will_status[TELOPT_NAWS] == WILL) {
    lines = output->get_size()->end - output->get_size()->start + 1;
    if(output->get_time_on())
      lines--;
    
    pos = 0;
    tmp[pos++] = IAC;
    tmp[pos++] = SB;
    tmp[pos++] = TELOPT_NAWS;
    
#define PUTSHORT(x) { if((tmp[pos++] = ((x)>>8)&0xff) == IAC) \
                        tmp[pos++] = IAC; \
			  if((tmp[pos++] = ((x))&0xff) == IAC) \
			    tmp[pos++] = IAC; }
    PUTSHORT(output->get_cols());
    PUTSHORT(lines);
    tmp[pos++] = IAC;
    tmp[pos++] = SE;
    
    if(write_socket(tmp, net, pos) < 0) {
      print_errmsg("write socket");
      rem_net();
      return(0);
    }
  }
  return(this);
}
#undef PUTSHORT
 
// ***************************************
//  check actions and process substitutes
// ***************************************

session *process_text(const uchar *line, uchar *dest, unsigned int &len,
		      session *ses)
{
  uchar buffer[BUFFER_SIZE], linebuf[BUFFER_SIZE], buf[BUFFER_SIZE];
  int lastsub = TRUE, anti_sub = FALSE;
  stack *st;
  listnode *ln;
  
  len = str_cat(dest, 0, BUFFER_SIZE, line);
  
  if(IS_SET(ses->ignore_flags, IGNORE_ACTION) &&
     IS_SET(ses->ignore_flags, IGNORE_SUBS))
    return(ses);
  
  st = new stack;
  
  for(ln = ses->actions; ln; ln = (listnode *)st->data) {
    ln->put_nextstack(st);
    if((ln->type == TYPE_ACTION && IS_SET(ses->ignore_flags, IGNORE_ACTION)) ||
       ((ln->type == TYPE_SUBS ||
	 ln->type == TYPE_GAGS ||
	 ln->type == TYPE_ANTI_SUB) &&
	(IS_SET(ses->ignore_flags, IGNORE_SUBS) || anti_sub)))
      continue;
    
    substitute_vars(ln->left, buffer, sizeof(buffer), FALSE, ses);
    if(ln->type == TYPE_ACTION || ln->type == TYPE_ANTI_SUB) {
      if(lastsub) {
	strip_ctrl_codes(linebuf, dest, 0);
	lastsub = FALSE;
      }
      if(check_one_action(linebuf, buffer, TRUE)) {
	if(ln->type == TYPE_ACTION) {
	  substitute_vars(ln->right, buffer, sizeof(buffer), TRUE, ses, FALSE);
	  if(IS_SET(ses->echo_flags, ECHO_ACTION)) {
	    sprintf((char *)buf, "[ACTION: %s]\n\r", buffer);
	    ses->print(buf);
	  }
	  if(!parse_input(buffer, ses)) {
	    delete st;
	    return(0);
	  }
	} else {
	  anti_sub = TRUE;
	}
      }
    } else {
      if(check_one_action(dest, buffer)) {
	lastsub = TRUE;
	
	if(ln->type == TYPE_GAGS) {
	  *dest = '\0';
	  len = 0;
	  delete st;
	  return(ses);
	} else
	  len = substitute_vars(ln->right, dest, BUFFER_SIZE, TRUE, ses);
      }
    }
  }
  delete st;
  return(ses);
}

/*******************************************************************/
/* read at most BUFFER_SIZE chars from mud - parse protecol stuff  */
/*******************************************************************/

int process_printing(session *ses)
{
  uchar tmpbuffer[BUFFER_SIZE], subresult[BUFFER_SIZE];
  uchar buffer[2*BUFFER_SIZE];
  uchar *endpos;

  unsigned int bufpos, pos;

  bufpos = 0;
  for(ses->netinput.get_line(tmpbuffer, sizeof(tmpbuffer));
      *tmpbuffer;
      ses->netinput.get_line(tmpbuffer, sizeof(tmpbuffer))) {

    for(endpos = tmpbuffer; *endpos; endpos++) {
      if(*endpos == '\n')
	break;
    }
    if(*endpos) {
      *endpos = '\0';

    } else {
      endpos = 0;
    }
    
    if(!process_text(tmpbuffer, subresult, pos, ses))
      return(-1);
    
    if(endpos && (*subresult || !*tmpbuffer)) {
      *endpos = '\n';
      pos = str_cat(subresult, pos, sizeof(subresult), endpos);
    }
    
    if(bufpos + pos < sizeof(buffer)) {
      bufpos = str_cat(buffer, bufpos, sizeof(buffer), subresult);
      
    } else {
      ses->print(buffer);
      ses->set_prompt((const uchar *)"");
      bufpos = str_cat(buffer, 0, sizeof(buffer), subresult);
    }
  }
  if(bufpos > 0) {
    ses->print(buffer);
    ses->set_prompt((const uchar *)"");
  }
  return(TRUE);
}

#define CUTLINE_FALSE 0
#define CUTLINE_BEGIN 1
#define CUTLINE_END   2

int parse_iac_escape_string(session *ses,
			    const uchar *orgstr, const uchar *orgend,
			    int &middle, int &got_ga)
{
  uchar letter;
  const uchar *str, *endpos;
  
  str    = orgstr;
  middle = TRUE;

  assert(*str == IAC && orgstr < orgend);
  str++;

  if(str < orgend) {
    
    letter = *str++;

    switch(letter) {
    case DONT:
    case DO:
    case WONT:
    case WILL:
      if(str < orgend) {
	middle = FALSE;
	
	if(!ses->do_telnet_protecol(letter, *str))
	  return(-1);
	str++;
      }
      break;
      
    case SB:

      endpos = str;
      while(endpos < orgend) {
	if(*endpos == IAC) {
	  endpos++;
	  
	  if(*endpos == IAC) {
	    endpos++;
	    
	  } else if(*endpos == SE) {
	    endpos++;
	    break;
	    
	  } else {
	    middle = FALSE;
	    return((endpos - 1) - orgstr); // skip over and ignore
	  }
	  
	} else {
	  endpos++;
	}
      }
      
      if(*(endpos - 2) == IAC && *(endpos - 1) == SE) {
	middle = FALSE;
	
	switch(*str) {
	case TELOPT_TTYPE:
	  if(parse_SB_TTYPE(*(str + 1), ses) < 0)
	    return(-1);
	  break;
	  
	case TELOPT_XDISPLOC:
	  if(parse_SB_XDISPLOC(*(str + 1), ses) < 0)
	    return(-1);
	  break;
	  
	case TELOPT_OLD_ENVIRON:
	case TELOPT_NEW_ENVIRON:
	  if(parse_SB_ENVIRON(str, endpos - 2, ses) < 0)
	    return(-1);
	  break;
	  
	default:
#ifdef TELCOM_DEBUG
	  un_sup_tel_opt(letter, *str, ses);
#endif
	  break;
	} /* switch */
	
	str = endpos;
      }
      break;
      
    case GA:
      got_ga = TRUE;
      middle = FALSE;
      break;
      
//    case EL:    /* erase the current line */
//    case EC:    /* erase the current character */
//    case AYT:   /* are you there */
//    case AO:    /* abort output--but let prog finish */

    case IP:    /* interrupt process--permanently */
      middle = FALSE;
      ses->print((const uchar *)"#Connection Teminated by foreing host.\n\r");
      ses->rem_net();
      prompt();
      return(-1);
      break;

//    case BREAK: /* break */
//    case DM:    /* data mark--for connect. cleaning */

    case NOP:   /* nop */
      middle = FALSE;
      break;

    case SE:    /* end sub negotiation */
      middle = FALSE;
      // hmm.. this should ever come here...
      break;
      
//    case EL:    /* erase the current line */
//    case EC:    /* erase the current character */
//    case AYT:   /* are you there */
//    case AO:    /* abort output--but let prog finish */
//    case BREAK: /* break */
//    case DM:    /* data mark--for connect. cleaning */
//    case EOR:   /* end of record (transparent mode) */
//    case ABORT: /* Abort process */
//    case SUSP:  /* Suspend process */
//    case xEOF:  /* End of file: EOF is already used... */
    default:
      middle = FALSE;
#ifdef TELCOM_DEBUG
      un_sup_tel_com(letter, ses);
#endif
      break;
    } /* switch(letter) */
  }
  
  return(str - orgstr);
}

int parse_esc_escape_string(const uchar *orgline,
			    uchar *buf, int &len, int &cutline, int &middle)
{
#ifdef DEBUG
  char errmsg[256];
#endif
  uchar num_buf[256], *num, letter;
  const uchar *line, *nums;
  int value;
  
  line    = orgline;
  len     = 0;
  cutline = FALSE;
  middle  = TRUE;
  
  assert(*line == ESC);
  
  buf[len++] = *line++;
  
  if(*line == '[') {  /* ansi/vt100 code */
    buf[len++] = *line++;

    if(*line == '?')
      buf[len++] = *line++;

    nums = line;
    while(isdigit(*line))
      buf[len++] = *line++;

    while(*line == ';') {
      buf[len++] = *line++;
      
      while(isdigit(*line))
	buf[len++] = *line++;
    }

    if(*line) {
      middle     = FALSE;
      letter     = *line++;
      buf[len++] = letter;
      
      switch(letter) {
      case 'H':
      case 'A':
      case 'B':
	cutline = TRUE;
	break;
	
      case 'n':
        /* don't return any string from these codes */
        len  = 0;
	*buf = '\0';

        while(*nums != 'n') {
	  
	  num = num_buf;
	  while(isdigit(*nums))
	    *num++ = *nums++;
	  *num = '\0';
	  
	  if(num == num_buf)
	    value = 0;
	  else
	    value = atoi((const char *)num_buf);
	  
	  switch(value) {
	  case 6:
	    *buf = 'R'; // lets return special mark... not nice coding but
	    len  = 1;   // really only way do it now.
	    break;
	    
	  default:
#ifdef DEBUG
	    sprintf(errmsg, "UNSUPORTED vt100-code ESC[%sn\n",
		    num_buf);
	    filelog(errmsg);
#endif
	    break;
	  } /* switch(value) */
	}
	break;
	
      default:
	break;
      } /* switch(letter) */

    } else {/* if(*line) */
      middle = TRUE;
    }
    
  } else { /* if(*line == '[') */
    
    if(*line) {
      middle  = FALSE;
      
      if(*line == 'D' || *line == 'M')
	cutline = TRUE;
      buf[len++] = *line++;
      
    } else {
      middle  = TRUE;
    }
  }
  buf[len] = '\0';
  
  return(line - orgline);
}

void ReadNet(session *ses)
{
  uchar readbuffer[512], tmpbuffer[BUFFER_SIZE];
  uchar linebuffer[BUFFER_SIZE], subresult[BUFFER_SIZE];
  uchar *rbuf, *dest, *str;
  int len, spos, middle, i, cutline, endlen, got_ga;
  unsigned int pos;
#ifdef X_DISPLAY_MISSING
  struct timeval null_time;
  fd_set input_set, exc_set;
  
  null_time.tv_sec = 0;
  null_time.tv_usec = 0;
#endif
  
  dest = linebuffer;
  rbuf = readbuffer;
  if(ses->netinput.middle_txt.len > 0) {
    middle = ses->netinput.middle_txt.is_middle();
    str = ses->netinput.middle_txt.text;
    
    if(middle) {
      while(str < ses->netinput.middle_txt.middle_pos)
	*dest++ = *str++;
      
      rbuf += str_cat(readbuffer, 0, sizeof(readbuffer), str);
      
    } else {
      dest += str_cat(linebuffer, 0, sizeof(linebuffer), str);
    }
    ses->netinput.middle_txt.clear();
  }
  *rbuf = '\0';
  *dest = '\0';
  
#ifdef X_DISPLAY_MISSING
  do {
#endif
    middle = FALSE;
    switch((len = read(ses->net, rbuf,
		       sizeof(readbuffer) - (rbuf - readbuffer) - 1))) {
    case 0:
      ses->print((const uchar *)"#Connection closed by foreing host.\n\r");
      ses->rem_net();
      prompt();
      return;
      
    case -1:
      ses->print_errmsg("Read from socket");
      ses->rem_net();
      prompt();
      return;

    default:
      *(rbuf + len) = '\0';
      
      if(ses->logfile) {
	fwrite(rbuf, len, sizeof(char), ses->logfile);
      }
      
      len += rbuf - readbuffer;
      rbuf = readbuffer;
      
      spos = 0;
      while(spos < len) {
	if(!readbuffer[spos]) { // ignore empty chars
	  spos++;
	  continue;
	}
	
	cutline = FALSE;
	got_ga  = FALSE;
	
	switch(readbuffer[spos]) {
	case ESC:
	  i = parse_esc_escape_string(readbuffer + spos, tmpbuffer,
				      endlen, cutline, middle);
	  
	  if(*tmpbuffer == 'R') { // special code ...
	    spos += i;
	    
	    if(dest != linebuffer) {
	      *dest = '\0';
	      ses->netinput.add_line(linebuffer);
	      dest = linebuffer;
	    }
	    if(process_printing(ses) < 0)
	      return;
	    
	    sprintf((char *)tmpbuffer, "%c[%d;%dR",
		    ESC,
		    (ses->get_output()->get_cursor()->row -
		     ses->get_output()->get_size()->start + 1),
		    ses->get_output()->get_cursor()->col);
	    if(write_socket(tmpbuffer, ses->get_socket(),
			    strlen((const char *)tmpbuffer)) < 0) {
	      ses->print_errmsg("write socket");
	      ses->rem_net();
	      return;
	    }
	  
	  } else if(*tmpbuffer) { /* if(*tmpbuffer == 'R') */
   	    if(middle) {
	      spos += i;
	      assert(spos == len);
	      
	      rbuf = readbuffer + str_cat(readbuffer, 0, sizeof(readbuffer),
					  tmpbuffer);
	      
	    } else { /* if(middle) */
	      if(cutline && dest != linebuffer) {
		*dest = '\0';
		ses->netinput.add_line(linebuffer);
		dest = linebuffer;
	      }
	      
	      spos += i;
	      dest += str_cat(dest, 0,
			      sizeof(linebuffer) - (dest - linebuffer),
			      tmpbuffer);
	    } /* if(middle) ... else */
	    
	  } else { /* if(*tmpbuffer == 'R') ... else if(*tmpbuffer) ... */
	    spos += i;
	  }
	  break;
	  
	case IAC:
	  if((i = parse_iac_escape_string(ses,
					  readbuffer + spos, readbuffer + len,
					  middle, got_ga)) < 0)
	    return;
	  
	  if(middle) {
	    assert(spos + i == len);
	    
	    rbuf = readbuffer;
	    while(spos < len)
	      *rbuf++ = readbuffer[spos++];
	    *rbuf = '\0';
	    
	  } else {
	    spos += i;
	  }
	  break;
	  
	case '\f':
	  while(readbuffer[spos] == '\f') {
	    *dest++ = '\f';
	    spos++;
	  }
	  *dest = '\0';
	  ses->netinput.add_line(linebuffer);
	  dest = linebuffer;
	  break;
	  
	case '\n':
	  *dest++ = '\n';
	  *dest++ = '\r';
	  spos++;
	  if(readbuffer[spos] == '\r') {
	    while(readbuffer[spos] == '\r')
	      spos++;
	  }
	  *dest = '\0';
	  ses->netinput.add_line(linebuffer);
	  dest = linebuffer;
	  break;
	  
	case '\r':
	  if(dest != linebuffer) {
	    *dest = '\0';
	    ses->netinput.add_line(linebuffer);
	    dest = linebuffer;
	  }
	  *dest++ = '\r';
	  
	  while(readbuffer[spos] == '\r')
	    spos++;
	  break;
	  
	default:
	  *dest++ = readbuffer[spos++];
	  break;
	}
      }
    } /* switch */
    
#ifdef X_DISPLAY_MISSING
    FD_ZERO(&input_set);
    FD_ZERO(&exc_set);
    FD_SET(ses->net, &input_set);
    FD_SET(ses->net, &exc_set);
    
    sigsetmask(signalmask);
    if(select(ses->net + 1, &input_set, (fd_set *)0, &exc_set, &null_time) < 0)
      syserr("select");
    sigsetmask(0);
  } while(FD_ISSET(ses->net, &input_set));
#endif

  if(process_printing(ses) < 0)
    return;

  *dest = '\0';
  *rbuf = '\0';
  if(middle || (ses->is_set_flag(SES_FLAG_GA_PROMPT) && !got_ga)) {
    len = dest - linebuffer;
    
    if(middle) {
      len += rbuf - readbuffer;
    }
    
    if(len > 0) {
      CREATE(ses->netinput.middle_txt.text, uchar, len + 1);
      pos = str_cat(ses->netinput.middle_txt.text, 0, len + 1,   linebuffer);
      
      if(middle) {
	ses->netinput.middle_txt.middle_pos = (ses->netinput.middle_txt.text +
					       pos);
	pos = str_cat(ses->netinput.middle_txt.text, pos, len + 1, readbuffer);
      }
      ses->netinput.middle_txt.len = pos;
    }

  } else if(*linebuffer) {
    
    if(!process_text(linebuffer, subresult, pos, ses))
      return;

    ses->print(subresult);
    if(ses->output != ses->input &&
       ses->input->get_active_ses() == ses &&
       !ses->input->get_hist_on()) {
      ses->input->print((const uchar *)"\r");
      ses->input->print(subresult);
      ses->input->print((const uchar *)"\033[K");
    }
    
    len = strlen((const char *)ses->promptbuf);
    
    str_cat(ses->promptbuf, len, MAX_PROMPT_SIZE, subresult);
    ses->prompt_pos = str_len(ses->promptbuf);
  }
  prompt();
}

/*****************************************************************/
/* respond according to the telnet protecol - weeeeelllllll..... */
/*****************************************************************/

session *session::do_telnet_protecol(int telcom, int telopt)
{
  /* we don't do anything here.. why should we? add the stuff yourself if
     you feel like being nice..... */
  
  unsigned char answer[20];
  int found = FALSE, needanswer = FALSE, size, lines;
  
  if(telopt < 0 || telopt >= NTELOPTS)
    return(this);

  answer[0] = IAC;
  answer[2] = telopt;
  answer[3] = '\0';
  
  switch(telcom) {
  case DONT:
    switch(telopt) {
    case TELOPT_ECHO:
      if(will_status[telopt] != WONT) {
	needanswer = TRUE;
	rem_flag(SES_FLAG_LINEMODE);
	will_status[telopt] = WONT;
      }
      found = TRUE;
      break;

    case TELOPT_SGA:
      if(will_status[telopt] != WILL) {
	needanswer = TRUE;
	will_status[telopt] = WILL;
      }
      found = TRUE;
      break;
      
    default:
      if(will_status[telopt] != WONT) {
	needanswer = TRUE;
	will_status[telopt] = WONT;
      }
      found = TRUE;
      break;
    }
    if(needanswer) {
      answer[1] = will_status[telopt];
      if(write_socket(answer, get_socket(), 3) < 0) {
	print_errmsg("write socket");
	rem_net();
	return(0);
      }
    }
    break;

  case DO:
    size = 3;
    switch(telopt) {
    case TELOPT_ECHO:
      if(will_status[telopt] != WILL) {
	needanswer = TRUE;
	set_flag(SES_FLAG_LINEMODE);
	will_status[telopt] = WILL;
      }
      found = TRUE;
      break;

    case TELOPT_SGA:
      if(will_status[telopt] != WILL) {
	needanswer = TRUE;
	will_status[telopt] = WILL;
      }
      found = TRUE;
      break;

    case TELOPT_STATUS:
      if(will_status[telopt] != WONT) {   // WILL
	needanswer = TRUE;
	will_status[telopt] = WONT;
      }
      found = TRUE;
      break;
      
    case TELOPT_TTYPE:
    case TELOPT_OLD_ENVIRON:
    case TELOPT_NEW_ENVIRON:
      if(will_status[telopt] != WILL) {
	needanswer = TRUE;
	will_status[telopt] = WILL;
      }
      found = TRUE;
      break;

    case TELOPT_XDISPLOC:
      if(will_status[telopt] != WILL &&
	 environment_list.find_value((const uchar *)"DISPLAY")) {
	needanswer = TRUE;
	will_status[telopt] = WILL;
	
      } else if(will_status[telopt] != WONT) {
	needanswer = TRUE;
	will_status[telopt] = WONT;
      }
      found = TRUE;
      break;
      
    case TELOPT_NAWS:
      if(will_status[telopt] != WILL) {
	needanswer = TRUE;
	will_status[telopt] = WILL;
	
	lines = output->get_size()->end - output->get_size()->start + 1;
	if(output->get_time_on())
	  lines--;
	
	answer[size++] = IAC;
	answer[size++] = SB;
	answer[size++] = TELOPT_NAWS;
	
#define PUTSHORT(x) { if((answer[size++] = ((x)>>8)&0xff) == IAC) \
			answer[size++] = IAC; \
			  if((answer[size++] = ((x))&0xff) == IAC) \
			    answer[size++] = IAC; }
	PUTSHORT(output->get_cols());
	PUTSHORT(lines);
#undef PUTSHORT
	answer[size++] = IAC;
	answer[size++] = SE;
      }
      found = TRUE;
      break;

    default:
      if(will_status[telopt] != WONT) {
	needanswer = TRUE;
	will_status[telopt] = WONT;
      }
      break;
    }
    
    if(needanswer) {
      answer[1] = will_status[telopt];
      if(write_socket(answer, get_socket(), size) < 0) {
	print_errmsg("write socket");
	rem_net();
	return(0);
      }
      if(answer[3] != '\0')
	answer[3] = '\0';
    }
    break;

  case WONT:
    switch(telopt) {
    case TELOPT_ECHO:
      if(do_status[telopt] != DONT) {
	set_flag(SES_FLAG_TERMECHO);
	do_status[telopt] = DONT;
      }
      found = TRUE;
      break;

    case TELOPT_SGA:
      if(do_status[telopt] != DONT) {
	do_status[telopt] = DONT;
      }
      found = TRUE;
      break;
      
    default:
      if(do_status[telopt] != DONT) {
	do_status[telopt] = DONT;
      }
      break;
    }
    answer[1] = do_status[telopt];
    if(write_socket(answer, get_socket(), 3) < 0) {
      print_errmsg("write socket");
      rem_net();
      return(0);
    }
    break;
    
  case WILL:
    switch(telopt) {
    case TELOPT_ECHO:
      if(do_status[telopt] != DO) {
	rem_flag(SES_FLAG_TERMECHO);
	window::get_active_win()->remove_user_text();
	do_status[telopt] = DO;
      }
      found = TRUE;
      break;
      
    case TELOPT_STATUS:
      if(do_status[telopt] != DONT) {
	do_status[telopt] = DONT;
      }
      found = TRUE;
      break;

    case TELOPT_SGA:
      if(do_status[telopt] != DO) {
	do_status[telopt] = DO;
      }
      found = TRUE;
      break;
      
    default:
      if(do_status[telopt] != DONT) {
	do_status[telopt] = DONT;
      }
      break;
    }
    answer[1] = do_status[telopt];
    if(write_socket(answer, get_socket(), 3) < 0) {
      print_errmsg("write socket");
      rem_net();
      return(0);
    }
    break;
  } /* switch */
  
#ifdef TELCOM_DEBUG
  if(!found) {
    un_sup_tel_opt(telcom, telopt, this);
  }
#endif
  return(this);
}
