/*********************************************************************/
/* File: screen.c - functions for the screen handling                */
/*                            HUT TINTIN                             */
/*                     coded by Ossi Tulonen                         */
/*********************************************************************/
/*
 * Copyright (c) 1992 1993 1994 1995 1996 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "structs.h"

#include <time.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "action.h"
#include "color.h"
#include "comms_help.h"
#include "environment.h"
#include "history.h"
#include "llist.h"
#include "main.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#include "utility.h"

#ifndef X_DISPLAY_MISSING
#  include <X11/IntrinsicP.h>
#  include <X11/Shell.h>
#  include <X11/Xatom.h>
#  include <X11/StringDefs.h>

extern XtAppContext MainContext;
extern unsigned long colors[];

#  include "xscreen.h"
#endif

#define INPUTBUF master_table.inputbuf
#define INPUTPOS master_table.inputbuf_pos
#define CURSOR   master_table.cursor_pos

basewindow *basewindow::basewins = 0;
window *window::active           = 0;

const uchar *CEL = 0;
const uchar *CSL = 0;
const uchar *CL  = 0;
const uchar *CES = 0;
const uchar *CSS = 0;
const uchar *CLEARSCREEN = 0;
const uchar *AUTOWARPON  = 0;
const uchar *AUTOWARPOFF = 0;
const uchar *HOME     = 0;
const uchar *CUP      = 0;
const uchar *CDOWN    = 0;
const uchar *FORWARD  = 0;
const uchar *BACKWARD = 0;
const uchar *RESETCOLOR = 0;
const uchar *MAXSCROLLSIZE = 0;
const uchar *INSERTLINE = 0;
const uchar *VBELL    = 0;
const uchar *MOVE     = 0;

int TABSIZE     = 0;

#ifdef HAVE_TERMIOS_H
struct termios original_state;
#else
struct sgttyb original_state;
struct tchars original_chars;
#endif

extern int default_chars_per_line;
extern int default_lines_in_scr;
extern struct master_table_data master_table;
extern unsigned short default_color;
extern struct key_func_table keys[256];

unsigned char *selection = 0;
int selection_len = 0;

#define CUR_STOP 5

// *******************************
//  class timeline_elem functions
// *******************************

timeline_elem::timeline_elem()
{
  def_line  = strdup(DEF_TIME_LINE);
  last_line = strdup((const uchar *)"");
  last_col  = strdup((const uchar *)"");
}

timeline_elem::~timeline_elem()
{
  if(def_line)
    free(def_line);
  
  if(last_line)
    free(last_line);
  
  if(last_col)
    free(last_col);
}

void timeline_elem::set(const uchar *newline)
{
  if(def_line)
    free(def_line);
  def_line = strdup(newline);
}

void timeline_elem::setlast(const uchar *newline)
{
  if(last_line)
    free(last_line);
  last_line = strdup(newline);
}

void timeline_elem::setlastcol(const uchar *newcol)
{
  if(last_col)
    free(last_col);
  last_col = strdup(newcol);
}

// **************************
//  class position functions
// **************************

position::position() : row(0), col(0)
{
}

// ****************************
// class winhist_elem functions
// ****************************

winhist_elem::winhist_elem()
{
  letter = (uchar)' ';
  colormode = default_color;
}

// ***********************
//  Base Window functions
// ***********************

basewindow::basewindow(const uchar *oname, const char *geometry,
		       const char *fontname, session *ses)
{
  colormode = default_color;
  
#ifndef X_DISPLAY_MISSING
  char buf[256];
  int i, def_x = 0, def_y = 0, res, n, font_width, font_height;
  unsigned int def_width = 80, def_height = 24;
  Arg ownargv[30];
  XEvent ev;
  
  if(MainDisplay) {
    sprintf(buf, "HUT TINTIN : %s", oname);
    
    font = GetFont(fontname, ses);
    font_width  = font->max_bounds.width;
    font_height = font->ascent + font->descent;
    
    n = 0;
    res = 0;
    XtSetArg(ownargv[n], XtNbackground, colors[DEFAULT_B]);
    n++;
    XtSetArg(ownargv[n], XtNborderWidth, 1);
    n++;
    
    if(geometry && *geometry) {
      res = XParseGeometry(geometry, &def_x, &def_y, &def_width, &def_height);
      if(def_width < 40)
	def_width = 40;
      if(def_height < 2)
	def_height = 2;
    }
    
    if(res & XValue) {
      if(res & XNegative)
	XtSetArg(ownargv[n], XtNx,
		 DisplayWidth(MainDisplay, DefaultScreen(MainDisplay)) -
		 (2 + font_width * def_width) + def_x);
      else
	XtSetArg(ownargv[n], XtNx, def_x);
      n++;
    }
    if(res & YValue) {
      if(res & YNegative)
	XtSetArg(ownargv[n], XtNy, 
		 DisplayHeight(MainDisplay, DefaultScreen(MainDisplay)) -
		 (2 + font_height * def_height) + def_y);
      else
	XtSetArg(ownargv[n], XtNy, def_y);
      n++;
    }
    if(res & XNegative && res & YNegative) {
      XtSetArg(ownargv[n], XtNwinGravity, SouthEastGravity);
      n++;
    } else if(res & XNegative) {
      XtSetArg(ownargv[n], XtNwinGravity, NorthEastGravity);
      n++;
    } else if(res & YNegative) {
      XtSetArg(ownargv[n], XtNwinGravity, SouthWestGravity);
      n++;
    }
    
    XtSetArg(ownargv[n], XtNwidth,    font_width  * def_width);
    n++;
    XtSetArg(ownargv[n], XtNminWidth, font_width * 40);
    n++;
    XtSetArg(ownargv[n], XtNwidthInc, font_width);
    n++;
    XtSetArg(ownargv[n], XtNheight,    font_height * def_height);
    n++;
    XtSetArg(ownargv[n], XtNminHeight, font_height * 2);
    n++;
    XtSetArg(ownargv[n], XtNheightInc, font_height);
    n++;
    XtSetArg(ownargv[n], XtNinput, TRUE); /* allow focus in  */
    n++;

    topLevelShellWidgetClass->core_class.expose = Redisplay;
    topLevelShellWidgetClass->core_class.resize = Resize;
    
    if(!(widget = XtAppCreateShell(buf, buf, topLevelShellWidgetClass,
				   MainDisplay, ownargv, n))) {
      if(ses) {
	ses->print((const uchar *)"Couldn't open window. (out of memory ?)\n\r");
	return;
      } else {
	fprintf(stderr, "Couldn't open window. (out of memory ?)\n");
	exit(1);
      }
    }

    XtAddEventHandler(widget, FocusChangeMask, TRUE,
		      HandleFocusChange, 0);
    XtAddEventHandler(widget, KeyPressMask, TRUE,
		      HandleKeyPress, 0);
    XtAddEventHandler(widget, ButtonPressMask, FALSE,
		      HandleMousePress, 0);
    XtAddEventHandler(widget, Button1MotionMask, FALSE,
		      HandleMouseMove, 0);
    XtAddEventHandler(widget, Button3MotionMask, FALSE,
		      HandleMouseMove, 0);
    
    /*
       XtAddEventHandler(widget, NoEventMask, TRUE,
       HandleSelection, this);
       */
    
    if(!(gc = XtGetGC(widget, 0, 0))) {
      XtDestroyWidget(widget);
      widget = 0;
      if(ses) {
	ses->print((const uchar *)"Couldn't open window graphic contect. (out of memory?)\n\r");
	return;
      } else {
	fprintf(stderr,
		"Couldn't open window graphic contect. (out of memory?)\n");
	exit(1);
      }
    }
    XSetGraphicsExposures(MainDisplay, gc, FALSE);
    
    XSetFont(MainDisplay, gc, font->fid);
    XSetForeground(MainDisplay, gc, colors[DEFAULT_F]);
    XSetBackground(MainDisplay, gc, colors[DEFAULT_B]);
    
    cols = widget->core.width/font_width;
    rows = widget->core.height/font_height;
    if((data = new winhist_elem *[rows])) {
      for(i = 0; i < rows; i++) {
	if(!(data[i] = new winhist_elem[cols + 1])) {
	  for(i -= 1;i >= 0; i--)
	    delete[] data[i];
	  delete[] data;
	  data = 0;
	  break;
	}
      }
    }
    if(!data) {
      XtReleaseGC(widget, gc);
      XtDestroyWidget(widget);
      widget = 0;
      if(ses) {
	ses->print((const uchar *)"#Out of memory.\n\r");
	return;
      } else {
	fprintf(stderr, "Not enought memory.\n");
	exit(1);
      }
    }
    h_col = 0;
    h_row = 0;
    clip_color = 0;
    XtRealizeWidget(widget);
    
    XWindowEvent(MainDisplay, XtWindow(widget), ExposureMask,
		 &ev);  /* lets wait here until window is open */
    XPutBackEvent(MainDisplay, &ev);

    x_wait_flush = FALSE;
  } else {
    widget = 0;
    gc = 0;
    font = 0;
    data = 0;
    h_col = 0;
    h_row = 0;
    clip_color = 0;
    x_wait_flush = FALSE;
#endif
    get_screen_size(cols, rows);
#ifndef X_DISPLAY_MISSING
  }
#endif
  Scroll.start = 1;
  Scroll.end   = rows;
  
  next = basewindow::basewins;
  basewindow::basewins = this;
  
  windows = new window(oname, this);
}

basewindow::~basewindow()
{

  basewindow *pre;
#ifndef X_DISPLAY_MISSING
  int i, not_in_list = FALSE;
  
  if(MainDisplay) {
    FreeFont(font);
    if(widget) {
      XtReleaseGC(widget, gc);
      XtDestroyWidget(widget);
      
      if(data) {
	for(i = 0; i < rows; i++)
	  delete[] data[i];
	delete[] data;
      }
    } else {
      not_in_list = TRUE;
    }
  }

  if(!not_in_list)
#endif
    if(this == basewindow::basewins) {
      basewindow::basewins = next;
    } else {
      for(pre = basewindow::basewins;pre && pre->next != this; pre = pre->next)
	;
      
      assert(pre && pre->next);
      pre->next = next;
    }

  if(!basewindow::basewins) {
#ifndef X_DISPLAY_MISSING
    if(MainDisplay)
      FreeFont(0);
#endif
    exit_tintin(0);
  }
}

#ifndef X_DISPLAY_MISSING
int basewindow::resize(const int nrows, const int ncols)
{
  int i, j, changed = FALSE;
  winhist_elem **olddata, *olddata2;
  
  assert(ncols > 0 && nrows > 0);
  
  if(ncols != cols) {
    for(i = 0; i < rows; i++) {
      olddata2 = data[i];
      if(!(data[i] = new winhist_elem[ncols + 1])) {
	delete[] olddata2;
	for(i -= 1; i >= 0; i--)
	  delete[] data[i];
	delete[] data;
	data = 0;
	delete this;
	
	master_table.main_session->print((const uchar *)"#ERROR out of memory.\n\r");
	return(-1);
      }
      for(j = 0; j < ncols && j < cols; j++) {
	data[i][j].letter = olddata2[j].letter;
	data[i][j].colormode = olddata2[j].colormode;
      }
      for(; j < ncols; j++) {
	data[i][j].colormode = data[i][j - 1].colormode;
      }
      delete[] olddata2;
    }
    cols = ncols;
    changed = TRUE;
  }
  
  if(nrows != rows) {
    olddata = data;
    if(!(data = new winhist_elem *[nrows])) {
      for(i = 0; i < rows; i++)
	delete[] olddata[i];
      delete[] olddata;
      data = 0;
      delete this;
      
      master_table.main_session->print((const uchar *)"#ERROR out of memory.\n\r");
      return(-1);
    }
    for(i = 1; i <= nrows; i++) {
      if(i <= rows)
	data[nrows - i] = olddata[rows - i];
      else {
	if(!(data[nrows - i] = new winhist_elem[ncols + 1])) {
	  for(j = i - 1; j > 0; j--)
	    delete[] data[nrows - j];
	  delete[] data;
	  data = 0;
	  for(i += 1; i <= rows; i++)
	    delete[] olddata[rows - i];
	  delete[] olddata;
	  delete this;
	  master_table.main_session->print((const uchar *)"#ERROR out of memory.\n\r");
	  return(-1);
	}
      }
    }
    for(; i <= rows; i++) {
      delete[] olddata[rows - i];
    }
    delete[] olddata;
    rows = nrows;
    changed = TRUE;
  }
  if(changed) {
    check_win_sizes();
    return(TRUE);
  }
  return(FALSE);
}
#endif

void basewindow::check_win_sizes()
{
  window *win, *nextwin;
  
  for(win = windows; win; win = nextwin) {
    nextwin = win->get_next();
    if(win->get_size()->end > rows) {
      if(rows - win->get_size()->start < 2) {
	delete win;
      }
    }
  }
  for(win = windows; win->get_next(); win = win->get_next())
    ;
  win->max_size();
}

// ********************
//  X-Window functions
// ********************

#ifndef X_DISPLAY_MISSING

void window::xprint(const uchar *str, int flush)
{
  int pcol, prow, len, font_width, font_height, font_ascent;
  
  assert(MainDisplay);
  
  if(!base->x_wait_flush) {
    font_width  = base->font->max_bounds.width;
    font_height = base->font->ascent + base->font->descent;
    font_ascent = base->font->ascent;
    
    if(base->colormode & COLBIT_NEGATIVE) {
      XSetBackground(MainDisplay, base->gc, colors[MODE_TO_T(base->colormode)]);
      XSetForeground(MainDisplay, base->gc, colors[MODE_TO_B(base->colormode)]);
    } else {
      XSetBackground(MainDisplay, base->gc, colors[MODE_TO_B(base->colormode)]);
      XSetForeground(MainDisplay, base->gc, colors[MODE_TO_T(base->colormode)]);
    }
    len = strlen((const char *)str);
    
    XDrawImageString(MainDisplay, XtWindow(base->widget), base->gc,
		     (cursor.col - 1) * font_width,
		     font_ascent + (cursor.row - 1) * font_height,
		     (const char *)str, len);
    
    if(base->colormode & COLBIT_BOLD)
      XDrawString(MainDisplay, XtWindow(base->widget), base->gc,
		  1 + (cursor.col - 1) * font_width,
		  font_ascent + (cursor.row - 1) * font_height,
		  (const char *)str, len);
    
    if(base->colormode & COLBIT_UNDERLINED) {
      XDrawLine(MainDisplay, XtWindow(base->widget), base->gc,
		(cursor.col - 1) * font_width,
		cursor.row * font_height - 1,
		(cursor.col - 1 + len) * font_width,
		cursor.row * font_height - 1);
    }
  }
  prow = cursor.row - 1;
  pcol = cursor.col - 1;
  
  while(*str && pcol < base->cols) {
    base->data[prow][pcol].colormode = base->colormode;
    base->data[prow][pcol++].letter = *str++;
  }
  if(pcol < base->cols)
    pcol++;
  
  cursor.col = pcol;
  
  if(flush)
    xprint_cursor();
}

void window::xprint_cursor(int all)
{
  basewindow *tmp_base;
  static position oldcursor;
  static basewindow *oldbase = 0;
  
  assert(MainDisplay);

  if(window::active != this)
    return;

  // let make sure oldbase is still there
  for(tmp_base = basewindow::get_bases();
      tmp_base;
      tmp_base = tmp_base->get_next())
    if(tmp_base == oldbase)
      break;
  
  if(tmp_base &&
     (oldcursor.col != cursor.col ||
      oldcursor.row != cursor.row ||
      oldbase != base ||
      all < 0) &&
     oldcursor.col > 0 &&
     oldcursor.row > 0 &&
     oldcursor.col <= oldbase->cols &&
     oldcursor.row <= oldbase->rows) {
    
    xprint_one(oldbase, oldcursor.col - 1, oldcursor.row - 1, FALSE);
    
    oldcursor.col = 0;
    oldcursor.row = 0;
    oldbase = 0;
  }
  
  if(all >= 0) {
    if(all > 0 ||
       oldcursor.col != cursor.col ||
       oldcursor.row != cursor.row ||
       oldbase != base) {
      
      oldcursor = cursor;
      oldbase = base;
      
      xprint_one(base, cursor.col - 1, cursor.row - 1, TRUE);
    }
  }
}

void window::xhist_scrollup()
{
  int i, j;
  struct winhist_elem *tmp;
  
  assert(MainDisplay);
  
  base->x_wait_flush = TRUE;
  
  tmp = base->data[base->Scroll.start - 1];
  
  for(i = base->Scroll.start; i < base->Scroll.end; i++)
    base->data[i - 1] = base->data[i];
  base->data[i - 1] = tmp;
  
  for(j = 0; j < base->cols; j++) {
    tmp[j].letter = (uchar)' ';
    tmp[j].colormode = base->colormode;
  }
}

void window::xhist_scrolldown()
{
  int i, j;
  struct winhist_elem *tmp;
  
  assert(MainDisplay);
  
  base->x_wait_flush = TRUE;
  
  tmp = base->data[base->Scroll.end - 1];
  
  for(i = base->Scroll.end - 1; i > base->Scroll.start - 1; i--)
    base->data[i] = base->data[i - 1];
  base->data[i] = tmp;
  
  for(j = 0; j < base->cols; j++) {
    tmp[j].letter = (uchar)' ';
    tmp[j].colormode = base->colormode;
  }
}

#endif

// **********************
// class window functions
// **********************

// first window in basewindow
window::window(const uchar *oname, basewindow *obase)
{
  session *mses;
  
  base = obase;
  next = 0;
  
  name = strdup(oname);
  colormode      = default_color;
  save_colormode = default_color;
  
  WinSize.start = 1;
  WinSize.end   = base->rows;
  Scroll.start  = 1;
  Scroll.end    = WinSize.end;
  cursor.col    = 1;
  cursor.row    = 1;
  timedelay     = -1;
  ptexrem       = FALSE;
  autowarp      = TRUE;
  window_hist_pos  = 0;
  window_hist_on   = 0;
  window_hist_clip = FALSE;
  
  active_ses   = 0;

  lastupdate = 0;

  window::active = this;

  mses = master_table.main_session;
  if(mses) {
    active_ses   = mses;
    if(mses->get_input()->get_active_ses() == mses)
      mses->get_input()->check_session();
    
    mses->set_input(this);
    mses->set_output(this);
  }

  clear();
}

window::window(const unsigned char *nam, const int size, window *win)
{
  window *pre;
  session *mses, *nextses;
  
  assert(size <= win->WinSize.end - win->WinSize.start - 2);
  
  base = win->base;

  name = strdup(nam);
  
  colormode      = default_color;
  save_colormode = default_color;
  
  WinSize.start      = win->WinSize.start;
  win->WinSize.start = WinSize.start + size;

  WinSize.end        = WinSize.start + size - 1;

  win->Scroll.start  = win->WinSize.start;
  
  Scroll.start = WinSize.start;
  Scroll.end   = WinSize.end;
  cursor.col   = 1;
  cursor.row   = WinSize.start;
  timedelay    = -1;
  ptexrem      = FALSE;
  autowarp     = TRUE;
  window_hist_pos = 0;
  window_hist_on  = 0;
  window_hist_clip = FALSE;

  mses = master_table.main_session;
  active_ses   = mses;
  if(mses->get_input()->get_active_ses() == mses)
    mses->get_input()->check_session();
  
  mses->set_input(this);
  mses->set_output(this);
  
  lastupdate = 0;
  
  if(base->windows == win) {
    next = base->windows;
    base->windows = this;
  } else {
    for(pre = base->windows; pre && pre->next != win; pre = pre->next)
      ;
    assert(pre);
    next = pre->next;
    pre->next = this;
  }
  window::active = this;

  redraw(TRUE);
  win->redraw();
  for(mses = session::get_sessions(); mses; mses = nextses) {
    nextses = mses->get_next();
    if(mses->get_output() == this || mses->get_output() == win) 
      mses->send_naws();
  }
}

window::~window()
{
  int tmp;
  window *pre = 0, *update;
  session *ses, *newactiv, *nextses;
  stack *st;
  
  for(st = stack::get_all(); st; st = st->get_next())
    if(this == (const window *)st->data)
      st->data = next;
  
  if(this == base->windows) {
    base->windows = next;
  } else {
    for(pre = base->windows; pre && pre->next != this; pre = pre->next)
      ;
    assert(pre);
    pre->next = next;
  }

  if(name)
    free(name);
  
  if(master_table.output_window == this)
    master_table.output_window = 0;
  
  if(!base->windows) {
    delete base;
  }
  
  update = 0;

  if(next) {
    next->WinSize.start = WinSize.start;
    next->Scroll.start  = WinSize.start;
    update = next;
  } else if(pre) {
    tmp = pre->WinSize.end - pre->Scroll.end;
    pre->WinSize.end = WinSize.end;
    pre->Scroll.end  = WinSize.end - tmp;
    update = pre;
  }
  
  if(window::active == this) {
    window::active = basewindow::get_bases()->windows;
    window::active->set_ptexrem();
    if(!window::active->active_ses) {
      for(newactiv = session::get_sessions();
	  newactiv;
	  newactiv = newactiv->get_next())
	if(newactiv->get_input() == window::active)
	  break;
      if(newactiv)
	window::active->active_ses = newactiv;
      else {
	if((newactiv = master_table.main_session)) {
	  if(newactiv->get_input()->get_active_ses() == newactiv)
	    newactiv->get_input()->check_session();
	  
	  window::active->active_ses = newactiv;
	  newactiv->set_input(window::active);
	}
      }
    }
#ifndef X_DISPLAY_MISSING
    if(MainDisplay && base != basewindow::get_bases())
      XMapRaised(MainDisplay, XtWindow(basewindow::get_bases()->widget));
#endif
  }

  for(ses = session::get_sessions(); ses; ses = nextses) {
    nextses = ses->get_next();
    
    if(ses->get_input() == this)
      ses->set_input(window::active);
    
    if(ses->get_output() == this) {
      ses->set_output(window::active);
      ses->send_naws();
    } else if(ses->get_output() == update)
      ses->send_naws();
  }
  
  if(master_table.main_session) {
    if(master_table.main_session->get_input() == this)
      master_table.main_session->set_input(window::active);
    
    if(master_table.main_session->get_output() == this)
      master_table.main_session->set_output(window::active);
  }
  
  if(update)
    update->redraw();
}

void window::move(const int row, const int col, const int flush)
{
  assert(row >= WinSize.start && row <= WinSize.end);
  assert(col >= 0 && col <= get_cols());
    
  if(base->Scrcursor.col != col || base->Scrcursor.row != row) {
#ifndef X_DISPLAY_MISSING
    if(!MainDisplay) {
#endif
#ifdef HAVE_LIBTERMCAP
      if(MOVE) {
	printf("%s", tgoto((const char *)MOVE, col - 1, row - 1));
      } else {
#endif
	printf("\033[%d;%dH", row, col);
#ifdef HAVE_LIBTERMCAP
      }
#endif
#ifndef X_DISPLAY_MISSING
    }
#endif
    
    base->Scrcursor.col = col;
    base->Scrcursor.row = row;
  }

  cursor.row = row;
  cursor.col = col;

  if(flush) {
#ifndef X_DISPLAY_MISSING
    if(MainDisplay)
      xprint_cursor();
    else
#endif
      fflush(stdout);
  }
}

void window::moveforward()
{
  check_color();

  if(++cursor.col > get_cols()) {
    if(autowarp) {
      cursor.col = 1;
      if(cursor.row < Scroll.end)
	cursor.row++;
      move(&cursor, TRUE);
    }
  } else {
    move(&cursor, TRUE);
  }
}

void window::movebackward()
{
  check_color();
  if(!(--cursor.col)) {
    cursor.col = get_cols();
    if(cursor.row > Scroll.start)
      cursor.row--;
    move(&cursor, TRUE);
  } else {
    move(&cursor, TRUE);
  }
}

void window::scroll_region(const int start, const int end)
{
  assert(start >= WinSize.start && end <= WinSize.end - get_time_on());
  
  if(start != base->Scroll.start || end != base->Scroll.end) {
#ifndef X_DISPLAY_MISSING
    if(!MainDisplay)
#endif
      printf("\033[%d;%dr", start, end);
    base->Scroll.start = start;
    base->Scroll.end = end;
    base->Scrcursor.col = -1;
    base->Scrcursor.row = -1;
  }
  move(&cursor, FALSE);
}

void window::max_scroll_region()
{
#ifndef X_DISPLAY_MISSING
  assert(!MainDisplay);
#endif
  
  printf("%s", MAXSCROLLSIZE);
  if(basewindow::get_bases()) {
    basewindow::get_bases()->Scroll.start = -1;
    basewindow::get_bases()->Scroll.end   = -1;
  }
}

void window::clear()
{
  int i, WinEnd;
#ifndef X_DISPLAY_MISSING
  int j, font_height;
#endif
  position pos;
  
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  check_color();
#ifndef X_DISPLAY_MISSING
  if(MainDisplay) {
    for(i = WinSize.start - 1; i < WinEnd; i++) {
      for(j = 0; j < get_cols(); j++) {
	base->data[i][j].letter = ' ';
	base->data[i][j].colormode = base->colormode;
      }
    }
    cursor.col = 1;
    cursor.row = WinSize.start;
    
    font_height = base->font->ascent + base->font->descent;
    
    XSetForeground(MainDisplay, base->gc, colors[MODE_TO_B(base->colormode)]);
    XFillRectangle(MainDisplay, XtWindow(base->widget),
		   base->gc,
		   0, (WinSize.start - 1) * font_height,
		   base->font->max_bounds.width  * get_cols(),
		   font_height * (WinEnd - WinSize.start + 1));
  } else {
#endif
    if(WinSize.start == 1 && WinEnd == get_rows()) {
      move(1, 1, TRUE);
      printf("%s", CLEARSCREEN);
    } else if(WinEnd == get_rows()) {
      move(WinSize.start, 1, TRUE);
      printf("%s", CES);
    } else {
      pos.row = WinSize.start;
      pos.col = 1;
      move(&pos, TRUE);
      
      printf("%s", CEL);
      
      if(WinSize.start != WinEnd) {
	for(i = WinSize.start + 1; i <= WinEnd; i++) {
	  move(i, 1, FALSE);
	  printf("%s", CEL);
	}
	move(&pos, TRUE);
      }
    }
    fflush(stdout);
#ifndef X_DISPLAY_MISSING
  }
#endif
  ptexrem = TRUE;
  if(window::active == this)
    master_table.utexrem = TRUE;
}

void window::clearend()
{
  int i, WinEnd;
#ifndef X_DISPLAY_MISSING
  int j, prow, pcol, font_width, font_height;
#endif
  position pos;
  
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  if(!window_hist_on) {

#ifndef X_DISPLAY_MISSING
    if(MainDisplay) {
      check_color();
      
      prow = cursor.row - 1;
      pcol = cursor.col - 1;
      
      font_width  = base->font->max_bounds.width;
      font_height = base->font->ascent + base->font->descent;
      
      XSetForeground(MainDisplay, base->gc, colors[MODE_TO_B(base->colormode)]);
      
      for(i = pcol; i < get_cols(); i++) {
	base->data[prow][i].letter = ' ';
	base->data[prow][i].colormode = base->colormode;
      }
      XFillRectangle(MainDisplay, XtWindow(base->widget),
		     base->gc,
		     pcol * font_width, prow * font_height,
		     font_width  * (get_cols() - pcol),
		     font_height);
      
      if(prow + 1 < WinEnd) {
	for(i = prow + 1; i < WinEnd; i++) {
	  for(j = 0; j < get_cols(); j++) {
	    base->data[i][j].letter = ' ';
	    base->data[i][j].colormode = base->colormode;
	  }
	}
	
	XFillRectangle(MainDisplay, XtWindow(base->widget),
		       base->gc,
		       0, (prow + 1) * font_height,
		       font_width  * get_cols(),
		       font_height * (WinEnd - cursor.row));
      }
      xprint_cursor(1);
    } else {
#endif
      move(&cursor, TRUE);
      
      check_color();
      
      if(WinEnd == get_rows()) {
	printf("%s", CES);
      } else {
	printf("%s", CEL);
	if(cursor.row != WinEnd) {
	  pos = cursor;
	  for(i = cursor.row + 1; i <= WinEnd; i++) {
	    move(i, 1, FALSE);
	    printf("%s", CEL);
	  }
	  move(&pos, TRUE);
	}
      }
      fflush(stdout);
#ifndef X_DISPLAY_MISSING
    }
#endif
  }
}

void window::clearstart()
{
  int i;
#ifndef X_DISPLAY_MISSING
  int j, prow, pcol, font_width, font_height;
#endif
  position pos;
  
  if(!window_hist_on) {
#ifndef X_DISPLAY_MISSING
    if(MainDisplay) {
      check_color();
      
      prow = cursor.row - 1;
      pcol = cursor.col - 1;
      
      font_width  = base->font->max_bounds.width;
      font_height = base->font->ascent + base->font->descent;
      
      XSetForeground(MainDisplay, base->gc, colors[MODE_TO_B(base->colormode)]);
      
      for(i = pcol; i >= 0; i--) {
	base->data[prow][i].letter = ' ';
	base->data[prow][i].colormode = base->colormode;
      }
      XFillRectangle(MainDisplay, XtWindow(base->widget),
		     base->gc,
		     0, prow * font_height,
		     font_width  * (pcol + 1),
		     font_height);
      
      if(prow >= WinSize.start) {
	for(i = prow - 1; i >= WinSize.start - 1; i--) {
	  for(j = 0; j < get_cols(); j++) {
	    base->data[i][j].letter = ' ';
	    base->data[i][j].colormode = base->colormode;
	  }
	}
	XFillRectangle(MainDisplay, XtWindow(base->widget),
		       base->gc,
		       0, (WinSize.start - 1) * font_height,
		       font_width  * get_cols(),
		       font_height * (cursor.row - WinSize.start));
      }
      xprint_cursor(1);
    } else {
#endif
      move(&cursor, TRUE);
      
      check_color();
      
      if(WinSize.start == 1) {
	printf("%s", CSS);
      } else {
	printf("%s", CSL);
	if(cursor.row > WinSize.start) {
	  pos = cursor;
	  for(i = cursor.row - 1; i >= WinSize.start; i--) {
	    move(i, 1, FALSE);
	    printf("%s", CEL);
	  }
	  move(&pos, TRUE);
	}
      }
      fflush(stdout);
#ifndef X_DISPLAY_MISSING
    }
#endif
  }
}

void window::clearline()
{
#ifndef X_DISPLAY_MISSING
  int i, font_width, font_height;
  
  if(MainDisplay) {
    font_width  = base->font->max_bounds.width;
    font_height = base->font->ascent + base->font->descent;
    
    for(i = 0; i < get_cols(); i++) {
      base->data[cursor.row - 1][i].letter = ' ';
      base->data[cursor.row - 1][i].colormode = base->colormode;
    }
    XSetForeground(MainDisplay, base->gc, colors[MODE_TO_B(base->colormode)]);
    
    XFillRectangle(MainDisplay, XtWindow(base->widget),
		   base->gc,
		   0, (cursor.row - 1) * font_height,
		   font_width * base->cols, font_height);
    xprint_cursor(1);
  } else
#endif
    printf("%s", CL);
}

void window::clearstartline()
{
#ifndef X_DISPLAY_MISSING
  int i, font_width, font_height;
  
  if(MainDisplay) {
    font_width  = base->font->max_bounds.width;
    font_height = base->font->ascent + base->font->descent;
    
    for(i = cursor.col - 1; i >= 0; i--) {
      base->data[cursor.row - 1][i].letter = ' ';
      base->data[cursor.row - 1][i].colormode = base->colormode;
    }
    XSetForeground(MainDisplay, base->gc, colors[MODE_TO_B(base->colormode)]);
    
    XFillRectangle(MainDisplay, XtWindow(base->widget),
		   base->gc,
		   0, (cursor.row - 1) * font_height,
		   font_width * (cursor.col), font_height);
    xprint_cursor(1);
  } else
#endif
    printf("%s", CSL);
}

void window::clearendline()
{
#ifndef X_DISPLAY_MISSING
  int i, font_width, font_height;
  
  if(MainDisplay) {
    font_width  = base->font->max_bounds.width;
    font_height = base->font->ascent + base->font->descent;
    
    for(i = cursor.col - 1; i < get_cols(); i++) {
      base->data[cursor.row - 1][i].letter = ' ';
      base->data[cursor.row - 1][i].colormode = base->colormode;
    }
    XSetForeground(MainDisplay, base->gc, colors[MODE_TO_B(base->colormode)]);
    
    XFillRectangle(MainDisplay, XtWindow(base->widget),
		   base->gc,
		   font_width * (cursor.col - 1) , (cursor.row - 1) * font_height,
		   font_width * (get_cols() - cursor.col + 1), font_height);
    xprint_cursor(1);
  } else
#endif
    printf("%s", CEL);
}

void window::redraw(const int log, const int position)
{
  uchar buf[2];
  int i = 0, num, pos, histpos, maxlines, startline, lines, tmp, line;
  session *ses = 0;
  window *win;
  const listnode *ln;
  struct position orgpos;
  basewindow *base;
  
  orgpos = cursor;
  clear();
  scroll_region(&Scroll);

  if(window_hist_on) {
    ses = window_hist_on;
    
    num = ses->get_hist_num();
    
    maxlines = WinSize.end - WinSize.start;
    if(get_time_on())
      maxlines--;
    startline = WinSize.start;
    
    if(num > maxlines)
      num = maxlines;
    
    lines = maxlines;
    if(lines > ses->get_hist_num())
      lines = ses->get_hist_num();

    tmp = (WinSize.start + (lines - 1) - orgpos.row);
    if(tmp < 0) {
      orgpos.col = 1;
      tmp = 0;
    }
    pos = window_hist_pos + tmp + 1;   // cursor position
    
    if(position < 0) {
      histpos = pos;
    } else if(position == 0) {
      histpos = pos + maxlines / 2;
    } else {
      histpos = pos + maxlines - 1;
    }

    if(histpos < num)
      histpos = num;
    
    if(histpos > ses->get_hist_num())
      histpos = ses->get_hist_num();
    
    orgpos.row = WinSize.start + histpos - pos;

    window_hist_pos = histpos - num;
    
    for(i = 0; i < num; i++) {
      move(startline + i, 1, FALSE);    /* cursor to right position */
      
      line = ses->get_hist_end() - histpos--;
      if(line < 0)
	line += ses->get_hist_size();
      
#ifndef X_DISPLAY_MISSING
      if(MainDisplay)
	xprint(ses->get_scrhistory(line));
      else
#endif
	printf("%s", ses->get_scrhistory(line));
    }
    
    if(ses->get_hist_markp() > window_hist_pos &&
       ses->get_hist_markp() <= window_hist_pos + num) {
      
      move(WinSize.start + window_hist_pos + num - ses->get_hist_markp(),
	   ses->get_hist_markc(), FALSE);
      
      if((ln = searchnode_list(ses->colors, (unsigned char *)"HISTMARK"))) {
	col_print(ln->right);
      } else {
	col_print(TIMELINEON);
      }
      
      line = ses->get_hist_end() - ses->get_hist_markp();
      if(line < 0)
	line += ses->get_hist_size();
      
      buf[0] = *(ses->get_scrhistory(line) + ses->get_hist_markc() - 1);
      buf[1] = '\0';
      col_print(buf);
    }
    
    cursor.row = orgpos.row;
    cursor.col = orgpos.col;
    print_win_hist_line(TRUE);
    ptexrem = FALSE;
  } else {
    if(active_ses && active_ses->get_output() == this)
      ses = active_ses;
    
    if(!ses) {
      for(base = basewindow::get_bases(); base && !ses; base = base->get_next()) {
	for(win = base->get_windows(); win; win = win->next) {
	  if(win->get_active_ses() &&
	     win->get_active_ses()->get_output() == this) {
	    ses = win->get_active_ses();
	    break;
	  }
	}
      }
    }
    
    if(!ses) {
      for(ses = master_table.main_session;
	  ses;
	  ses = ses->get_next())
	if(ses->get_output() == this)
	  break;
    }

    if(ses) {
      num = ses->get_hist_num();

      maxlines = Scroll.end - Scroll.start;
      startline = Scroll.start;
      
      if(num > maxlines)
	num = maxlines;
      
      histpos = window_hist_pos + num;
      
      if(histpos > ses->get_hist_num())
	histpos = ses->get_hist_num();
      
      for(i = 0; i < num; i++) {
	move(startline + i, 1, FALSE);    /* cursor to right position */
	
	pos = ses->get_hist_end() - histpos--;
	if(pos < 0)
	  pos += ses->get_hist_size();
	
#ifndef X_DISPLAY_MISSING
	if(MainDisplay)
	  xprint(ses->get_scrhistory(pos));
	else
#endif
	  printf("%s", ses->get_scrhistory(pos));
      }
    }
    move(Scroll.start + i, 1, TRUE);    /* cursor to right position */
    print_prompt(TRUE, log);
  }
  
  print_time(TRUE);
}

void window::bell()
{
#ifndef X_DISPLAY_MISSING
  int revert_to_return;
  XWindowAttributes winattr;
  Window focus_return;
  XEvent ev;
#endif
  
  print_time(FALSE, TRUE);
  
#ifndef X_DISPLAY_MISSING
  if(MainDisplay) {
    if(XGetWindowAttributes(MainDisplay, XtWindow(base->widget), &winattr)) {
      if(winattr.map_state == IsViewable) {
	XRaiseWindow(MainDisplay, XtWindow(base->widget));
      } else {
	XGetInputFocus(MainDisplay, &focus_return, &revert_to_return);
	XMapRaised(MainDisplay, XtWindow(base->widget));
	
	/* lets wait until get focus in event */
	XWindowEvent(MainDisplay, XtWindow(base->widget),
		     FocusChangeMask, &ev);

	/* now we just have to hope old focused window is still there... */
	if(XGetWindowAttributes(MainDisplay, focus_return, &winattr) &&
	   winattr.map_state == IsViewable) {
	  XSetInputFocus(MainDisplay, focus_return, revert_to_return,
			 CurrentTime);
	}
      }
    } else {
      XMapRaised(MainDisplay, XtWindow(base->widget));
    }
    if(master_table.is_set_flag(MASTER_FLAG_VISUAL_BELL))
      XVBell(this);

    if(master_table.is_set_flag(MASTER_FLAG_BELL))
      XBell(MainDisplay, 0);
    
  } else {
#endif
    if(VBELL && (master_table.is_set_flag(MASTER_FLAG_VISUAL_BELL)))
      printf("%s", VBELL);
    
    if(master_table.is_set_flag(MASTER_FLAG_BELL))
      putchar('\a');
    
    fflush(stdout);
#ifndef X_DISPLAY_MISSING
  }
#endif
}

void window::print(const unsigned char *cptr, const unsigned char *add)
{
#ifdef DEBUG
  char errmsg[256];
#endif
  unsigned char buf[BUFFER_SIZE + 100], *str, tmpbuf[100], *tmp;
  unsigned char numbuf[100], *num, numbuf2[100], *num2, color[BUFFER_SIZE];
  unsigned short tmp_colmode;
  int showlen = 0, addlen = 0, i, j, decmode, WinEnd, value;
  position orgcursor;
#ifndef X_DISPLAY_MISSING
  struct winhist_elem *eline;
#endif
  static position save_cursor;

  scroll_region(&Scroll);
  move(&cursor, TRUE);
  
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  if(add) {
    addlen = str_len(add);
    if(addlen >= get_cols())
      return;
    
    while(*cptr && (*cptr == '\r' || *cptr == '\n'))
      cptr++;
  }
  
  str = buf;
  while(*cptr) {
    showlen = cursor.col - 1;
    if(add) {
#ifndef X_DISPLAY_MISSING
      if(MainDisplay)
	xprint(add);
      else
#endif
	printf("%s", add);
      showlen += addlen;
      cursor.col = addlen + 1;
    }
    while(*cptr &&
	  (showlen < get_cols() || !autowarp || (*cptr < ' ' || *cptr == 127))) {
      if(*cptr == ESC) {
	*str = '\0';
#ifndef X_DISPLAY_MISSING
	if(MainDisplay)
	  xprint(buf, TRUE);
	else
#endif
	  printf("%s", buf);
	cursor.col = showlen + 1;
	if(cursor.col > get_cols())
	  cursor.col = get_cols();
	
	str = buf;
	
	tmp = tmpbuf;
	*tmp++ = *cptr++;
	if(*cptr == '[') {
	  *tmp++ = *cptr++;
	  
	  decmode = FALSE;
	  if(*cptr == '?') {
	    decmode = TRUE;
	    *tmp++ = *cptr++;
	  }
	  
	  num = numbuf;
	  while(isdigit(*cptr)) {
	    *num++ = *cptr;
	    *tmp++ = *cptr++;
	  }
	  *num = '\0';
	  
	  num2 = numbuf2;
	  if(*cptr == ';') {
	    *tmp++ = *cptr++;
	    while(isdigit(*cptr)) {
	      *num2++ = *cptr;
	      *tmp++  = *cptr++;
	    }
	  }
	  *num2 = '\0';
	  while(*cptr == ';') {
	    *tmp++ = *cptr++;
	    while(isdigit(*cptr)) {
	      *num2  = *cptr;
	      *tmp++ = *cptr++;
	    }
	  }
	  *tmp++ = *cptr;
	  *tmp = '\0';
	  
	  if(decmode && *cptr != 'J' && *cptr != 'K') {
	    switch(*cptr) {
	    case 'l':
	      value = 0;
	      if(*numbuf) {
		value = atoi((const char *)numbuf);
	      }
	      
	      switch(value) {
	      case 0:                   // IGNORED
		break;
	      case 7:                   // AUTOWARP OFF
		autowarp = FALSE;
		break;
	      }
#ifdef DEBUG
	      if(*numbuf2) {
		filelog("UNSUPORTED multi vt100-code ESC[?;;l");
	      }
#endif
	      cptr++;
	      break;
	    case 'h':
	      value = 0;
	      if(*numbuf) {
		value = atoi((const char *)numbuf);
	      }
	      
	      switch(value) {
	      case 0:                   // IGNORED
		break;
	      case 7:                   // AUTOWARP ON
		autowarp = TRUE;
		break;
	      }
#ifdef DEBUG
	      if(*numbuf2) {
		filelog("UNSUPORTED multi vt100-code ESC[;;h");
	      }
#endif
	      cptr++;
	      break;
	    default:
#ifdef DEBUG
	      *tmpbuf = 'C';
	      sprintf(errmsg, "UNSUPORTED ansi-code ES%s", tmpbuf);
	      filelog(errmsg);
#endif
	      cptr++;
	      break;
	    }  /* switch */
	  } else {
	    switch(*cptr) {
	    case 'm':
	      tmp_colmode = base->colormode;
	      color_to_bits(tmpbuf, base->colormode, this);
#ifndef X_DISPLAY_MISSING
	      if(MainDisplay)
		base->data[cursor.row-1][showlen].colormode = base->colormode;
	      else
#endif
		if(bits_to_color(tmp_colmode, base->colormode, color) > 0)
		  printf("%s", color);
	      
	      colormode = base->colormode;
	      cptr++;
	      break;
	    case 'r':
	      if(!add) {
		value = 1;
		if(*numbuf) {
		  value = atoi((const char *)numbuf);
		  if(value < 1)
		    value = 1;
		  else if(value > WinEnd - WinSize.start + 1)
		    value = WinEnd - WinSize.start + 1;
		}
		
		if(*numbuf2) {
		  j = atoi((const char *)numbuf2);
		  if(j < 1)
		    j = 1;
		  else if(j > WinEnd - WinSize.start + 1)
		    j = WinEnd - WinSize.start + 1;
		} else
		  j = WinEnd - WinSize.start + 1;
		
		Scroll.start = WinSize.start + value - 1;
		Scroll.end   = WinSize.start + j - 1;
		
		scroll_region(&Scroll);
	      }
	      cptr++;
	      break;
	    case 'A':
	      if(!add) {
		value = 1;
		if(*numbuf) {
		  value = atoi((const char *)numbuf);
		}
		
		if(cursor.row - value < WinSize.start)
		  value = cursor.row - WinSize.start;
		
		if(value > 0) {
		  cursor.row -= value;
#ifndef X_DISPLAY_MISSING
		  if(!MainDisplay) {
#endif
		    if(value == 1)
		      printf("%s", CUP);
		    else
		      printf("\033[%dA", value);
#ifndef X_DISPLAY_MISSING
		  }
#endif
		  base->Scrcursor = cursor;
		}
	      }
	      cptr++;
	      break;
	    case 'B':
	      if(!add) {
		value = 1;
		if(*numbuf) {
		  value = atoi((const char *)numbuf);
		}
		
		if(cursor.row + value > WinEnd)
		  value = WinEnd - cursor.row;
		
		if(value > 0) {
		  cursor.row += value;
#ifndef X_DISPLAY_MISSING
		  if(!MainDisplay) {
#endif
		    if(value == 1)
		      printf("%s", CDOWN);
		    else
		      printf("\033[%dB", value);
#ifndef X_DISPLAY_MISSING
		  }
#endif
		  base->Scrcursor = cursor;
		}
	      }
	      cptr++;
	      break;
	    case 'C':
	      if(!add) {
		value = 1;
		if(*numbuf) {
		  value = atoi((const char *)numbuf);
		}
		
		if(showlen + value >= get_cols())
		  value = get_cols() - showlen - 1;
		
		if(value > 0) {
		  showlen += value;
		  cursor.col += value;
#ifndef X_DISPLAY_MISSING
		  if(!MainDisplay) {
#endif
		    if(value == 1)
		      printf("%s", FORWARD);
		    else
		      printf("\033[%dC", value);
#ifndef X_DISPLAY_MISSING
		  }
#endif
		  base->Scrcursor = cursor;
		}
	      }
	      cptr++;
	      break;
	    case 'D':
	      if(!add) {
		value = 1;
		if(*numbuf) {
		  value = atoi((const char *)numbuf);
		}
		
		if(showlen - value < 0)
		  value = showlen;
		
		if(value > 0) {
		  showlen -= value;
		  cursor.col -= value;
#ifndef X_DISPLAY_MISSING
		  if(!MainDisplay) {
#endif
		    if(value == 1)
		      printf("%s", BACKWARD);
		    else
		      printf("\033[%dD", value);
#ifndef X_DISPLAY_MISSING
		  }
#endif
		  base->Scrcursor = cursor;
		}
	      }
	      cptr++;
	      break;
	    case 'H':
	      if(!add) {
		if(*numbuf) {
		  i = atoi((const char *)numbuf);
		  if(i <= 0)
		    i = 1;
		  else if(i > WinEnd - WinSize.start + 1)
		    i = WinEnd - WinSize.start + 1;
		} else
		  i = 1;
		
		if(*numbuf2) {
		  j = atoi((const char *)numbuf2);
		  if(j <= 0)
		    j = 1;
		  else if(j > get_cols())
		    j = get_cols();
		} else
		  j = 1;
		
		move(WinSize.start + i - 1, j, TRUE);
		showlen = cursor.col - 1;
	      }
	      cptr++;
	      break;
	    case 'J':
	      if(!add) {
		if(!*numbuf || (*numbuf == '0' && !*(numbuf + 1))) {
		  clearend();
		} else if(*numbuf == '1' && !*(numbuf + 1)) {
		  clearstart();
		} else if(*numbuf == '2' && !*(numbuf + 1)) {
		  orgcursor = cursor;
		  clear();
		  
		  move(&orgcursor, TRUE);
		  showlen = cursor.col - 1;
		} else {
#ifdef DEBUG
		  *tmpbuf = 'C';
		  sprintf(errmsg, "UNSUPORTED ansi-code ES%s", tmpbuf);
		  filelog(errmsg);
#endif
		}
	      }
	      cptr++;
	      break;
	    case 'K':
	      if(!add) {
		value = 0;
		if(!*numbuf || (value = atoi((const char *)numbuf) == 0)) {
		  clearendline();
		} else if(value == 1) {
		  clearstartline();
		} else if(value == 2) {
		  clearline();
		} else {
#ifdef DEBUG
		  *tmpbuf = 'C';
		  sprintf(errmsg, "UNSUPORTED ansi-code ES%s", tmpbuf);
		  filelog(errmsg);
#endif
		}
	      }
	      cptr++;
	      break;
	    case 'L':
	      if(!add) {
#ifndef X_DISPLAY_MISSING
		if(MainDisplay) {
		  i = 1;
		  if(*numbuf)
		    i = atoi((const char *)numbuf);

		  base->Scroll.start = cursor.row;
		  while(i > 0) {
		    xhist_scrolldown();
		    i--;
		  }
		  scroll_region(&Scroll);
		} else {
#endif
		  if(*numbuf) {
		    printf("\033[%sL", numbuf);
		  } else {
		    printf("%s", INSERTLINE);
		  }
#ifndef X_DISPLAY_MISSING
		}
#endif
	      }
	      cptr++;
	      break;
	    case 'M':
	      if(!add) {
#ifndef X_DISPLAY_MISSING
		if(MainDisplay) {
		  i = 1;
		  if(*numbuf)
		    i = atoi((const char *)numbuf);
		  
		  base->Scroll.start = cursor.row;
		  while(i > 0) {
		    xhist_scrollup();
		    i--;
		  }
		  scroll_region(&Scroll);
		} else {
#endif
		  if(*numbuf)
		    printf("\033[%sM", numbuf);
		  else
		    printf("\033[M");
#ifndef X_DISPLAY_MISSING
		}
#endif
	      }
	      cptr++;
	      break;
	    case 'P':
	      if(!add) {
#ifndef X_DISPLAY_MISSING
		if(MainDisplay) {
		  i = 1;
		  if(*numbuf)
		    i = atoi((const char *)numbuf);
		  
		  eline = base->data[cursor.row - 1];
		  for(j = cursor.col - 1; j + i < base->cols; j++) {
		    eline[j].letter    = eline[j + i].letter;
		    eline[j].colormode = eline[j + i].colormode;
		  }
		  for(; j < base->cols; j++) {
		    eline[j].letter    = (uchar)' ';
		    eline[j].colormode = base->colormode;
		  }
		  xprint_many(base, cursor.col - 1, cursor.row - 1,
			      base->cols - cursor.col + 1, FALSE);
		  if(window::get_active_win() == this)
		    xprint_cursor(TRUE);
		} else {
#endif
		  if(*numbuf)
		    printf("\033[%sP", numbuf);
		  else
		    printf("\033[P");
#ifndef X_DISPLAY_MISSING
		}
#endif
	      }
	      cptr++;
	      break;
	    case '@':
	      if(!add) {
#ifndef X_DISPLAY_MISSING
		if(MainDisplay) {
		  i = 1;
		  if(*numbuf)
		    i = atoi((const char *)numbuf);
		  
		  eline = base->data[cursor.row - 1];

		  for(j = base->cols - 1; j - i >= cursor.col - 1; j--) {
		    eline[j].letter    = eline[j - i].letter;
		    eline[j].colormode = eline[j - i].colormode;
		  }
		  for(; j >= cursor.col - 1; j--) {
		    eline[j].letter    = (uchar)' ';
		    eline[j].colormode = base->colormode;
		  }
		  
		  xprint_many(base, cursor.col - 1, cursor.row - 1,
			      base->cols - cursor.col + 1, FALSE);
		  if(window::get_active_win() == this)
		    xprint_cursor(TRUE);
		} else {
#endif
		  if(*numbuf)
		    printf("\033[%s@", numbuf);
		  else
		    printf("\033[@");
#ifndef X_DISPLAY_MISSING
		}
#endif
	      }
	      cptr++;
	      break;
	    default:
#ifdef DEBUG
	      *tmpbuf = 'C';
	      sprintf(errmsg, "UNSUPORTED ansi-code ES%s", tmpbuf);
	      filelog(errmsg);
#endif
	      cptr++;
	      break;
	    }  /* switch */
	  }
 	} else {
	  if(!add) {
	    switch(*cptr) {
	    case '=':                // numeric keypad application mode
#ifndef X_DISPLAY_MISSING
	      if(!MainDisplay)
#endif
		printf("\033=");
	      break;
	    case '>':                // numeric keypad numeric mode
#ifndef X_DISPLAY_MISSING
	      if(!MainDisplay)
#endif
		printf("\033>");
	      break;
	    case '7':                // should do other things to but we won't
	      save_cursor = cursor;
	      break;
	    case '8':
	      if(save_cursor.row > 0 && save_cursor.col > 0) {
		move(&save_cursor, TRUE);
		showlen = cursor.col - 1;
	      }
	      break;
	    case 'D':
#ifndef X_DISPLAY_MISSING
	      if(MainDisplay) {
		if(cursor.row < Scroll.end) {
		  cursor.row++;
		} else
		  xhist_scrollup();
	      } else {
#endif
		printf("\033D");
		if(cursor.row < Scroll.end) {
		  cursor.row++;
		}
#ifndef X_DISPLAY_MISSING
	      }
#endif
	      base->Scrcursor = cursor;
	      break;
	    case 'M':
#ifndef X_DISPLAY_MISSING
	      if(MainDisplay) {
		if(cursor.row > Scroll.start) {
		  cursor.row--;
		} else
		  xhist_scrolldown();
	      } else {
#endif
		printf("\033M");
		if(cursor.row > Scroll.start) {
		  cursor.row--;
		}
#ifndef X_DISPLAY_MISSING
	      }
#endif
	      base->Scrcursor = cursor;
	      break;
	    default:
#ifdef DEBUG
	      sprintf(errmsg, "UNSUPORTED ESC code ESC%c", *cptr);
	      filelog(errmsg);
#endif
	      break;
	    } /* switch */
	  }
	  if(*cptr)
	    cptr++;
	}
      } else {
	if(*cptr >= ' ' && *cptr != 127) {  // 127 = delete
	  if(showlen < get_cols())
	    showlen++;
	  *str++ = *cptr++;
	} else {
	  *str = '\0';
#ifndef X_DISPLAY_MISSING
	  if(MainDisplay)
	    xprint(buf, TRUE);
	  else {
#endif
	    printf("%s", buf);
	    cursor.col = showlen + 1;
	    if(cursor.col > get_cols())
	      cursor.col = get_cols();
#ifndef X_DISPLAY_MISSING
	  }
#endif
	  str = buf;
	  
	  switch(*cptr) {
	  case '\b':                       // backspace
	    if((!add && showlen > 0) ||
	       (add && (showlen - addlen) > 0)) {
#ifndef X_DISPLAY_MISSING
	      if(!MainDisplay)
#endif
		printf("\b");
	      
	      showlen--;
	      cursor.col--;
	    }
	    cptr++;
	    break;
	  case '\t':                       // tab
	    showlen++;
	    if(showlen % TABSIZE)
	      showlen += TABSIZE - (showlen % TABSIZE);
	    
	    if(showlen + 1 > get_cols())
	      showlen = get_cols() - 1;
	    
	    move(cursor.row, showlen + 1, TRUE);
	    cptr++;
	    break;
	  case '\a':                       // bell
	    bell();
	    cptr++;
	    break;
	  case '\r':
	    if(!add) {
	      showlen = 0;
	      cursor.col = 1;
#ifndef X_DISPLAY_MISSING
	      if(!MainDisplay)
#endif
		*str++ = '\r';
	    }
	    cptr++;
	    break;
	  case '\f':
	    if(!add) {
#ifndef X_DISPLAY_MISSING
	      if(MainDisplay)
		if(cursor.row >= Scroll.end)
		  xhist_scrollup();
		else {
		  cursor.row++;
		}
	      else
#endif
		if(cursor.row < Scroll.end) {
		  cursor.row++;
		}
	      base->Scrcursor = cursor;
	    }
	    cptr++;
	    break;
	  case '\n':
	    if(add) {
#ifndef X_DISPLAY_MISSING
	      if(MainDisplay) {
		if(cursor.row < Scroll.end)
		  cursor.row++;
		else
		  xhist_scrollup();
		cursor.col = 1;
	      } else {
#endif
		printf("\n\r");
		if(cursor.row < Scroll.end)
		  cursor.row++;
		cursor.col = 1;
#ifndef X_DISPLAY_MISSING
	      }
#endif
	      base->Scrcursor = cursor;
	      showlen = 0;
	      
	      while(*cptr && (*cptr == '\n' || *cptr == '\r'))
		cptr++;
	      
	      if(*cptr) {
#ifndef X_DISPLAY_MISSING
		if(MainDisplay)
		  xprint(add);
		else
#endif
		  printf("%s", add);
		showlen += addlen;
		cursor.col = addlen + 1;
	      }
	    } else {
#ifndef X_DISPLAY_MISSING
	      if(MainDisplay) {
		if(cursor.row < Scroll.end)
		  cursor.row++;
		else
		  xhist_scrollup();
	      } else {
#endif
		printf("\n");
		if(cursor.row < Scroll.end)
		  cursor.row++;
#ifndef X_DISPLAY_MISSING
	      }
#endif
	      base->Scrcursor = cursor;
	      cptr++;
	    }
	    str = buf;
	    break;
	  case '\017':        // ctrl-o
#ifndef X_DISPLAY_MISSING
	    if(!MainDisplay)
#endif
	      printf("%c", *cptr);
	    cptr++;
	    break;
	  default:
	    sprintf((char *)numbuf, "%3.3o", (int)*cptr);
	    showlen += 4;
	    if(showlen > get_cols())
	      showlen = get_cols();
	    *str++ = '\\';
	    *str++ = numbuf[0];
	    *str++ = numbuf[1];
	    *str++ = numbuf[2];
	    cptr++;
	    break;
	  }
	}
      }
    }
    *str = '\0';
    
    if((showlen == get_cols() && autowarp) ||
       (add && (*buf || cursor.col > 1))) {
      
#ifndef X_DISPLAY_MISSING
      if(MainDisplay) {
	if(*buf)
	  xprint(buf);
	if(cursor.row < Scroll.end)
	  cursor.row++;
	else
	  xhist_scrollup();
	cursor.col = 1;
      } else {
#endif
	if(*buf)
	  printf("%s\n\r", buf);
	else
	  printf("\n\r");
	
	if(cursor.row < Scroll.end)
	  cursor.row++;
	cursor.col = 1;
#ifndef X_DISPLAY_MISSING
      }
#endif
      base->Scrcursor = cursor;
    } else {
#ifndef X_DISPLAY_MISSING
      if(MainDisplay) {
	xprint(buf, TRUE);
	cursor.col = showlen + 1;
	if(cursor.col > get_cols())
	  cursor.col = get_cols();
      } else {
#endif
	printf("%s", buf);
	cursor.col = showlen + 1;
	if(cursor.col > get_cols())
	  cursor.col = get_cols();
#ifndef X_DISPLAY_MISSING
      }
#endif
      base->Scrcursor = cursor;
    }
    str = buf;
  }
#ifndef X_DISPLAY_MISSING
  if(!MainDisplay)
#endif
    fflush(stdout);
}

void window::col_print(const uchar *str)
{
  char errmsg[256];
  uchar buffer[BUFFER_SIZE], *buf, tmpbuf[MAX_LINE_LENGTH], *tmp;
  uchar numbuf[MAX_LINE_LENGTH], *num, color[BUFFER_SIZE];
  unsigned short tmp_colmode;
  int len, value;
#ifndef X_DISPLAY_MISSING
  int prow, pcol;
#endif
  
  buf = buffer;
  len = 0;
  while(*str) {
    if(*str == ESC) {
      *buf = '\0';
      if(len > 0) {
#ifndef X_DISPLAY_MISSING
	if(MainDisplay) {
	  xprint(buffer);
	} else {
#endif
	  printf("%s", buffer);
	  cursor.col += len;
	  if(cursor.col > get_cols())
	    cursor.col = get_cols();
#ifndef X_DISPLAY_MISSING
	}
#endif
	base->Scrcursor = cursor;
	buf = buffer;
	len = 0;
      }
      
      tmp = tmpbuf;
      *tmp++ = *str++;
      if(*str == '[') {
	*tmp++ = *str++;
	
	num = numbuf;
	while(isdigit(*str)) {
	  *num++ = *str;
	  *tmp++ = *str++;
	}
	*num = '\0';
	
	while(*str == ';') {
	  *tmp++ = *str++;
	  while(isdigit(*str))
	    *tmp++ = *str++;
	}
	*tmp++ = *str;
	*tmp = '\0';
	
	switch(*str) {
	case 'm':
	  tmp_colmode = base->colormode;
	  color_to_bits(tmpbuf, base->colormode, this);
#ifndef X_DISPLAY_MISSING
	  if(MainDisplay) {
	    prow = cursor.row - 1;
	    pcol = cursor.col - 1;
	    
	    base->data[prow][pcol].colormode = base->colormode;
	  } else {
#endif
	    if(bits_to_color(tmp_colmode, base->colormode, color) > 0)
	      printf("%s", color);
#ifndef X_DISPLAY_MISSING
	  }
#endif
	  str++;
	  break;
	case 'K':
	  value = 0;
	  if(!*numbuf || (value = atoi((const char *)numbuf) == 0)) {
	    clearendline();
	  } else if(value == 1) {
	    clearstartline();
	  } else if(value == 2) {
	    clearline();
	  } else {
#ifdef DEBUG
	    *tmpbuf = 'C';
	    sprintf(errmsg, "UNSUPORTED ansi-code ES%s", tmpbuf);
	    filelog(errmsg);
#endif
	  }
	  str++;
	  break;
	default:
	  sprintf(errmsg, "BUg BUg Bug 1 Wrong kind text col_print() %s%c",
		  tmpbuf + 2, *str);
	  filelog(errmsg);
	  str++;
	  break;
	}
      } else {
	str++;
	filelog("BUg BUg Bug 2 Wrong kind text col_print()");
      }
    } else {
      if(*str >= ' ' && *str != 127) {
	*buf++ = *str++;
	len++;
      } else {
	str++;
	filelog("BUg BUg Bug 3 Wrong kind text col_print()");
      }
    }
  }
  *buf = '\0';

  if(*buffer) {
#ifndef X_DISPLAY_MISSING
    if(MainDisplay) {
      xprint(buffer);
    } else {
#endif
      printf("%s", buffer);
      cursor.col += len;
      if(cursor.col > get_cols())
	cursor.col = get_cols();
#ifndef X_DISPLAY_MISSING
    }
#endif
    base->Scrcursor = cursor;
  }
}

int window::change_time(const int delay, const uchar *newtimeline)
{
  unsigned char buf[BUFFER_SIZE];
  session *ses, *nextses;
  
  if(newtimeline) {
    timeline.set(newtimeline);
  }
  
  if(delay < 0) {
    if(timedelay < 0) {
      return(FALSE);
    } else {
      active_ses->print((const uchar *)"#Time line OFF.\n\r");
      timedelay = delay;
      if(WinSize.end == Scroll.end + 1)
	Scroll.end++;
      
      redraw(TRUE);
      for(ses = session::get_sessions(); ses; ses = nextses) {
	nextses = ses->get_next();
	if(ses->get_output() == this)
	  ses->send_naws();
      }
    }
  } else {
    sprintf((char *)buf, "#Time line ON. (updating every %d seconds.)\n\r", delay);
    if(timedelay < 0) {
      if(Scroll.start + 1 < Scroll.end ||
	 Scroll.end < WinSize.end) {
	active_ses->print(buf);
	timedelay = delay;
	if(WinSize.end == Scroll.end)
	  Scroll.end--;
	
	redraw(TRUE);
	for(ses = session::get_sessions(); ses; ses = nextses) {
	  nextses = ses->get_next();
	  if(ses->get_output() == this)
	    ses->send_naws();
	}
      } else {
	active_ses->print((const uchar *)"#Too small window to time line.\n\r");
	return(FALSE);
      }
    } else {
      active_ses->print(buf);
      if((timedelay != 0 && delay == 0) ||
	 (timedelay == 0 && delay != 0)) {
	timedelay = delay;
	print_time(TRUE);
      } else
	timedelay = delay;
    }
  }
  return(TRUE);
}

void window::print_time(const int org_all, const int bell)
{
  unsigned char buf[256];
  uchar tmp_buf1[BUFFER_SIZE], tmp_buf2[BUFFER_SIZE];
  uchar begin[MAX_LINE_LENGTH], end[MAX_LINE_LENGTH], *str;
  const uchar *orgline, *line, *orgcol, *colstr;
  int len, pos, two_parts, hours, mins;
  time_t timep, secs;
  struct position orgpos;
  struct tm *timedata;
  const listnode *ln;

  window *activ;
  
  int all = org_all;
  
  if(!(orgline = timeline.get_defline()) || timedelay < 0)
    return;
  
  // getting time info
  timep = time(0);
  timedata = localtime(&timep);
  
  colstr = TIMELINEON; // color to print time line
  
  if(active_ses) {
    if((ln = searchnode_list(active_ses->colors,
			     (const uchar *)"TIMELINEON"))) {
      colstr = ln->right;
    }
    secs = timep - active_ses->get_time();
    substitute_vars(orgline, tmp_buf1, sizeof(tmp_buf1),
		    TRUE, active_ses, FALSE);
  } else {
    secs = timep - master_table.main_session->get_time();
    substitute_vars(orgline, tmp_buf1, sizeof(tmp_buf1),
		    TRUE, master_table.main_session, FALSE);
  }

  // calculate online time on active/main session
  hours = secs  / (60*60);
  secs -= hours * (60*60);
  mins  = secs  / 60;
  secs -= mins  * 60;
  strip_ctrl_codes(tmp_buf2, tmp_buf1, 0);
  
  line = tmp_buf2;
  str  = begin;
  pos  = 0;
  two_parts = FALSE;
  
  while(*line) {
    if(*line == '%') {
      line++;
      
      assert(*line);
      
      switch(*line++) {
      case '%':
	str[pos++] = '%';
	break;
	
      case 'a':  // is active window
	if(this == window::active) {
	  pos = str_cat(str, pos, MAX_LINE_LENGTH,
			(const uchar *)" ACTIVE ");
	}
	break;
	
      case 'b':  // bell
	if(bell) {
	  pos = str_cat(str, pos, MAX_LINE_LENGTH,
			(const uchar *)" <BELL> ");
	}
	break;

      case 'd':  // date
	sprintf((char *)buf, "%2.2d.%2.2d.19%2.2d",
		timedata->tm_mday,
                timedata->tm_mon + 1,
                timedata->tm_year);
	pos = str_cat(str, pos, MAX_LINE_LENGTH, buf);
	break;
	
      case 'f':
	str[pos] = '\0';
	str = end;
	pos = 0;
	two_parts = TRUE;
	break;
	
      case 'o':  // sessions online time
	sprintf((char *)buf, "%2.2d:%2.2d:%2.2d", hours, mins, (int)secs);
	pos = str_cat(str, pos, MAX_LINE_LENGTH, buf);
	break;
	
      case 's':  // active session name
	if(active_ses) {
	  pos = str_cat(str, pos, MAX_LINE_LENGTH, active_ses->get_name());
	  
	} else {
	  pos = str_cat(str, pos, MAX_LINE_LENGTH, (const uchar *)"<none>");
	}
	break;
	
      case 't':  // time
	if(timedelay > 0) {
	  sprintf((char *)buf, "%2.2d:%2.2d:%2.2d",
		  timedata->tm_hour,
		  timedata->tm_min,
		  timedata->tm_sec);
	  pos = str_cat(str, pos, MAX_LINE_LENGTH, buf);
	}
	break;

      case 'T':  // ticker time to left next tick
	if(active_ses && active_ses->is_set_flag(SES_FLAG_TICKER)) {
	  sprintf((char *)buf, "%d", active_ses->sec_to_tick);
	  pos = str_cat(str, pos, MAX_LINE_LENGTH, buf);
	  
	} else {
	  pos = str_cat(str, pos, MAX_LINE_LENGTH, (const uchar *)"-1");
	}
	break;
	
      case 'w':  // window name
	pos = str_cat(str, pos, MAX_LINE_LENGTH, name);
	break;
	
      default:
	abort();
	break;
      } /* switch */
      
    } else {
      str[pos++] = *line++;
    }
  }

  if(two_parts) {
    
    len = strlen((const char *)begin);
    
    while(len < (get_cols() - pos))
      begin[len++] = ' ';
    
    pos = 0;
    while(len < get_cols())
      begin[len++] = end[pos++];
    
  } else {
    
    while(pos < get_cols())
      begin[pos++] = ' ';
  }
  begin[get_cols()] = '\0';
  
  // check if same as last time
  orgline = timeline.get_lastline();
  orgcol  = timeline.get_lastcol();

  if(strcmp((const char *)colstr, (const char *)orgcol)) {// color has changed
    all = TRUE;
    timeline.setlastcol(colstr);
  }
  
  if(!strcmp((const char *)begin,  (const char *)orgline) && !all)
    return;
  
  orgpos = cursor;
  
  if(this != window::active)
    move(&cursor, FALSE);
  
  check_color();
  
  if(all) {
    move(WinSize.end, 1, FALSE);
    col_print(colstr);
    
#ifndef X_DISPLAY_MISSING
    if(MainDisplay) {
      xprint(begin);
    } else
#endif
      printf("%s", begin);
    
  } else {
    
    for(pos = 0; pos < get_cols(); pos++) {
      if(begin[pos] != orgline[pos])
	break;
    }
    
    move(WinSize.end, pos + 1, FALSE);
    col_print(colstr);
    
    do {
      
      len = 0;
      while(pos < get_cols()) {
	if(begin[pos] == orgline[pos] &&
	   (pos + 2 >= get_cols() ||
	    (begin[pos + 1] == orgline[pos + 1] &&
	     begin[pos + 2] == orgline[pos + 2]))) {
	  break;
	}
	end[len++] = begin[pos++];
      }
      end[len] = '\0';
      
#ifndef X_DISPLAY_MISSING
      if(MainDisplay) {
	xprint(end);
      } else
#endif
	printf("%s", end);
      
      
      while(pos < get_cols()) {
	if(begin[pos] != orgline[pos]) {
	  move(WinSize.end, pos + 1, FALSE);
	  break;
	}
	
	pos++;
      }
    } while(pos < get_cols());
  }
  
  cursor = orgpos;
  activ = window::get_active_win();
  activ->move(activ->get_cursor(), TRUE);
  activ->check_color();
  
#ifndef X_DISPLAY_MISSING
  if(!MainDisplay)
#endif
    fflush(stdout);
  
  timeline.setlast(begin);
}

void window::print_win_hist_line(const int all, unsigned char *text)
{
  unsigned char buffer[20], buf2[BUFFER_SIZE];
  int tmp, lines, num, WinEnd, i;
  static int lastkey = 0;
  position orgpos;
  
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  check_color();
  
  if(window_hist_on) {
    orgpos = cursor;

    lines = WinEnd - WinSize.start;
    
    for(i = 0; i < 'z' - 'a' + 1; i++)
      if(keys[i + 1].func == func_exit_win_hist)
	break;

    if(i < 'z' - 'a' + 1) {
      sprintf((char *)buffer, "ctrl-%c", i + 'a');
    } else {
      sprintf((char *)buffer, "no key");
    }

    num = window_hist_on->get_hist_num();
    tmp = ((window_hist_on->get_hist_size() > num ? num :
	    window_hist_on->get_hist_size()) -
	   (lines > num ? num : lines));
    
    if((master_table.inputbuf_pos && window::active == this) || text) {
      if(all || lastkey != i) {
	lastkey = i;
	move(WinEnd, 1, FALSE);               /* last line begin */
	
	if(text) {
	  sprintf((char *)buf2, "[%s to exit] (pos: %3d/%3d) ERROR: %s%s",
		  buffer,
		  window_hist_pos, tmp, text, CEL);
	  col_print(buf2);
	} else {
	  sprintf((char *)buf2, "[%s to exit] (pos: %3d/%3d) S: %s%s",
		  buffer,
		  window_hist_pos, tmp, master_table.inputbuf, CEL);
	  col_print(buf2);
	}
      } else {
	move(WinEnd, 24, FALSE);              /* last line begin */
	if(text) {
	  sprintf((char *)buf2, "%3d/%3d) ERROR: %s%s",
		  window_hist_pos, tmp, text, CEL);
	  col_print(buf2);
	} else {
	  sprintf((char *)buf2, "%3d/%3d) S: %s%s",
		  window_hist_pos, tmp, master_table.inputbuf, CEL);
	  col_print(buf2);
	}
      }
    } else {
      if(all || lastkey != i) {
	lastkey = i;
	move(WinEnd, 1, FALSE);               /* last line begin */
	sprintf((char *)buf2, "[%s to exit] (pos: %3d/%3d)%s",
		buffer,
		window_hist_pos, tmp, CEL);
	col_print(buf2);
      } else {
	move(WinEnd, 24, FALSE);              /* last line begin */
	sprintf((char *)buf2, "%3d/%3d)%s", window_hist_pos, tmp, CEL);
	col_print(buf2);
      }
    }
    move(&orgpos, TRUE);
    fflush(stdout);
  }
}

void window::print_prompt(const int all, const int log)
{
  int lines, pos, prompt_len = 0;
  session *ses = 0;
  window *win = 0;
  basewindow *base;
  
  check_color();

  if(active_ses) {
    prompt_len = active_ses->get_prompt_len();
  } else {
    for(base = basewindow::get_bases(); base && !ses; base = base->get_next()) {
      for(win = base->get_windows(); win; win = win->next) {
	if(win->get_active_ses() &&
	   win->get_active_ses()->get_output() == this) {
	  ses = win->get_active_ses();
	  break;
	}
      }
    }
    if(!ses) {
      for(ses = master_table.main_session;
	  ses;
	  ses = ses->get_next())
	if(ses->get_output() == this)
	  break;
    }
    if(ses)
      prompt_len = ses->get_prompt_len();
    else
      prompt_len = 2;
  }
  
  if(all && ptexrem) {
    if(active_ses) {
      if(active_ses->is_set_flag(SES_FLAG_LINEMODE)) {
	if(log) {
	  active_ses->print(active_ses->get_prompt());
	  if(this != active_ses->get_output())
	    print(active_ses->get_prompt());
	} else
	  print(active_ses->get_prompt());
      }
    } else if(!log) {
      if(ses) {
	if(ses->is_set_flag(SES_FLAG_LINEMODE))
	  print(ses->get_prompt());
      } else
	print((const uchar *)"> ");
    }
    ptexrem = FALSE;
  }

  if(this == window::active && active_ses) {
    if(master_table.utexrem) {
      if(INPUTPOS) {
	if(!window_hist_on && active_ses->is_set_flag(SES_FLAG_TERMECHO)) {
	  print(INPUTBUF);
	  
	  if(INPUTPOS > CURSOR) {
	    lines = ((prompt_len + INPUTPOS) / get_cols()) -
	      ((prompt_len + CURSOR) / get_cols());
	    
	    pos = (prompt_len + CURSOR) % get_cols();
	    
	    move(cursor.row - lines, pos + 1, TRUE);
	  }
	}
      }
      master_table.utexrem = FALSE;
    }
  }

  fflush(stdout);
}

void window::remove_user_text()
{
  int lines, p_lines, p_col;
  int prompt_len;
  prompt_len = active_ses ? active_ses->get_prompt_len() : 0;
  
  check_color();
  
  if(this == window::active) {
    if(!window_hist_on) {
      if(INPUTPOS && !master_table.utexrem) {
	if(CURSOR) {
	  p_lines = prompt_len / get_cols();
	  p_col   = prompt_len - (p_lines * get_cols()) + 1;
	  lines = ((prompt_len + CURSOR) / get_cols()) - p_lines;
	  
	  if(cursor.row - lines < WinSize.start)
	    lines = cursor.row - WinSize.start;
	  
	  move(cursor.row - lines, p_col, FALSE);
	} else {
	  move(&cursor, FALSE);
	}
	clearend();
      }
    }
    master_table.utexrem = TRUE;
  }
}

void window::hist_movebackward()
{
  int lines, WinEnd, line, len, tmp, pos;

  check_color();
  
  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  lines = WinEnd - WinSize.start;
  
  if(lines > window_hist_on->get_hist_num())
    lines = window_hist_on->get_hist_num();
  
  tmp = WinSize.start + (lines - 1) - cursor.row;
  if(tmp < 0)
    tmp = 0;
  
  pos = window_hist_pos + tmp + 1;

  if(cursor.col > 1) {
    line = window_hist_on->get_hist_end() - pos;
    if(line < 0)
      line += window_hist_on->get_hist_size();
    
    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    
    cursor.col--;
    if(cursor.col > len)
      cursor.col = len;
    
    move(&cursor, TRUE);
  } else {
    if(pos < window_hist_on->get_hist_num()) {
      pos++;
      line = window_hist_on->get_hist_end() - pos;
      if(line < 0)
	line += window_hist_on->get_hist_size();
      
      cursor.col = strlen((const char *)window_hist_on->get_scrhistory(line));
      
      if(cursor.row > WinSize.start) {
	cursor.row--;
	move(&cursor, TRUE);
      } else {
	hist_backward(1);
      }
    }
  }
}

void window::hist_moveforward()
{
  int lines, WinEnd, line, len, tmp, pos;

  WinEnd = WinSize.end;
  if(get_time_on())
    WinEnd--;
  
  lines = WinEnd - WinSize.start;
  
  if(lines > window_hist_on->get_hist_num())
    lines = window_hist_on->get_hist_num();
  
  tmp = WinSize.start + (lines - 1) - cursor.row;
  if(tmp < 0)
    tmp = 0;
  pos = window_hist_pos + tmp + 1;
  
  line = window_hist_on->get_hist_end() - pos;
  if(line < 0)
    line += window_hist_on->get_hist_size();
  
  len = strlen((const char *)window_hist_on->get_scrhistory(line));
  if(len > window::get_cols())
    len = window::get_cols();

  check_color();

  if(cursor.col < len) {
    cursor.col++;
    
    move(&cursor, TRUE);
  } else {
    if(pos > 0) {
      cursor.col = 1;
      
      if(cursor.row < WinSize.start + (lines - 1)) {
	cursor.row++;
	move(&cursor, TRUE);
      } else {
	hist_forward(1);
      }
    }
  }
}

void window::hist_backward(int amount)
{
  int lines, pos, WinEnd, num;
  position orgpos;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;

    lines = WinEnd - WinSize.start;
    
    if(cursor.row > WinSize.start) {
      num = cursor.row - WinSize.start;
      if(num > amount)
	num = amount;
      
      cursor.row -= num;
      move(&cursor, TRUE);
      amount -= num;
    }
    
    if(window_hist_pos + lines < window_hist_on->get_hist_num() && amount){
      
      if(window_hist_pos + lines + amount > window_hist_on->get_hist_num())
	amount = window_hist_on->get_hist_num() - window_hist_pos - lines;
      
      orgpos = cursor;
      
      if(amount < lines) {
	scroll_region(WinSize.start, WinEnd - 1);
	
	check_color();
	
	for(num = 0; num < amount; num++) {
	  window_hist_pos++;
	  pos = (window_hist_on->get_hist_end() -
		 (window_hist_pos + lines));
	  if(pos < 0)
	    pos += window_hist_on->get_hist_size();
	  
	  /* 1 line down all lines */
	  move(WinSize.start, 1, FALSE);
#ifndef X_DISPLAY_MISSING
	  if(MainDisplay) {
	    xhist_scrolldown();
	    xprint(window_hist_on->get_scrhistory(pos));
	  } else {
#endif
	    printf("%s%s", INSERTLINE,window_hist_on->get_scrhistory(pos));
	    base->Scrcursor.col = -1; /* force cursor move */
#ifndef X_DISPLAY_MISSING
	  }
#endif
	}
	cursor = orgpos;
	print_win_hist_line(FALSE);
      } else {
	window_hist_pos += amount;
	redraw(FALSE, -1);
      }
    }
  }
}

void window::hist_forward(int amount)
{
  int pos, WinEnd, num, lines;
  position orgpos;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;

    lines = WinEnd - 1 - WinSize.start;
    if(lines > window_hist_on->get_hist_num() - 1)
      lines = window_hist_on->get_hist_num() - 1;
    
    if(cursor.row < WinSize.start + lines) {
      num = (WinSize.start + lines) - cursor.row;
      if(num > amount)
	num = amount;
      
      cursor.row += num;
      move(&cursor, TRUE);
      amount -= num;
    } else if(window_hist_pos == 0) {
      exit_win_history();
    }
    
    if(window_hist_pos > 0 && amount) {
      if(window_hist_pos < amount)
	amount = window_hist_pos;
      
      orgpos = cursor;
      
      if(amount < WinEnd - WinSize.start) {
	scroll_region(WinSize.start, WinEnd - 1);
	
	move(WinEnd - 1, get_cols(), FALSE);
	
	check_color();
	
	for(num = 0; num < amount; num++) {
	  pos = window_hist_on->get_hist_end() - window_hist_pos;
	  if(pos < 0)
	    pos += window_hist_on->get_hist_size();
	  window_hist_pos--;


#ifndef X_DISPLAY_MISSING
	  if(MainDisplay) {
	    xhist_scrollup();
	    cursor.col = 1;
	    cursor.row = WinEnd - 1;
	    xprint(window_hist_on->get_scrhistory(pos));
	  } else
#endif
	    printf("\n\r%s", window_hist_on->get_scrhistory(pos));
	}
	cursor = orgpos;
	print_win_hist_line(FALSE);
      } else {
	window_hist_pos -= amount;
	redraw(FALSE, 1);
      }
    }
  }
}

void window::hist_home()
{
  int WinEnd, pos;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;
    
    WinEnd--;
    pos = WinEnd - WinSize.start;
    if(pos > window_hist_on->get_hist_num() - 1)
      pos = window_hist_on->get_hist_num() - 1;
    
    if(window_hist_pos) {
      cursor.row = WinSize.start + pos;
      cursor.col = 1;
      
      window_hist_pos = 0;
      redraw();
    } else {
      move(WinSize.start + pos, 1, TRUE);
    }
  }
}

void window::hist_end()
{
  int lines, pos = 0, WinEnd, line, len;
  
  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;

    lines = WinEnd - WinSize.start;
    pos = window_hist_on->get_hist_num() - lines;
    if(pos < 0)
      pos = 0;
    
    line = window_hist_on->get_hist_end() - window_hist_on->get_hist_num();
    if(line < 0)
      line += window_hist_on->get_hist_size();
    
    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    if(len > window::get_cols())
      len = window::get_cols();
    
    if(window_hist_pos != pos) {
      cursor.row = WinSize.start;
      cursor.col = len;
      window_hist_pos = pos;
      redraw();
    } else {
      move(WinSize.start, len, TRUE);
    }
  }
}

void window::hist_pageup()
{
  int lines, WinEnd;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;
    
    lines = WinEnd - WinSize.start;
    
    hist_backward(lines);
  }
}

void window::hist_pagedown()
{
  int lines, WinEnd;
  
  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;
    
    lines = WinEnd - WinSize.start;
    
    hist_forward(lines);
  }
}

int window::hist_search(const unsigned char *str, const int forward)
{
  int lines, endpos, pos, found, num, i, line, WinEnd, tmp, col;

  if(window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;
    
    lines = WinEnd - WinSize.start;
    
    if(lines > window_hist_on->get_hist_num())
      lines = window_hist_on->get_hist_num();
    
    tmp = WinSize.start + (lines - 1) - cursor.row;
    if(tmp < 0)
      tmp = 0;
    
    pos = window_hist_pos + tmp + 1;
    col = cursor.col - 1 + forward;
    
    endpos = window_hist_on->get_hist_num();
    
    found = FALSE;
    for(num = 0; pos <= endpos; pos++, num++) {
      line = window_hist_on->get_hist_end() - pos;
      if(line < 0)
	line += window_hist_on->get_hist_size();
      
      if(col > 0) {
	if(col > (tmp = strlen((const char *)window_hist_on->get_scrhistory(line)))) {
	  col = 0;
	  continue;
	}
      }
      
      for(i = col; *(window_hist_on->get_scrhistory(line) + i) != '\0' ; i++)
	if(is_abrev(window_hist_on->get_scrhistory(line) + i, str)) {
	  found = TRUE;
	  col = i + 1;
	  break;
	}
      if(found)
	break;
      
      col = 0;
    }
    if(!found) {
      print_win_hist_line(FALSE, (unsigned char *)"Couldn't find such string");
      return(FALSE);
    }
    
    if(col > window::get_cols())
      col = window::get_cols();
    
    cursor.col = col;
    hist_backward(num);
  }
  return(TRUE);
}

void window::hist_set_mark()
{
  uchar buf[2];
  int lines, num, line, pos, len;
  position orgpos;
  const listnode *ln;
  
  if(window_hist_on) {
    num = window_hist_on->get_hist_num();
    
    lines = WinSize.end - WinSize.start;
    if(get_time_on())
      lines--;
    
    if(num > lines)
      num = lines;

    orgpos = cursor;

    if(window_hist_on->get_hist_markp() > window_hist_pos &&
       window_hist_on->get_hist_markp() <= window_hist_pos + num) {
      
      move(WinSize.start + window_hist_pos + num -
	   window_hist_on->get_hist_markp(),
	   window_hist_on->get_hist_markc(), FALSE);
      
      check_color();
      
      line = window_hist_on->get_hist_end() - window_hist_on->get_hist_markp();
      if(line < 0)
	line += window_hist_on->get_hist_size();
      
      buf[0] = *(window_hist_on->get_scrhistory(line) +
		 window_hist_on->get_hist_markc() - 1);
      buf[1] = '\0';
#ifndef X_DISPLAY_MISSING
      if(MainDisplay)
	xprint(buf);
      else
#endif
	printf("%s", buf);
      move(&orgpos, FALSE);
    }
    if((ln = searchnode_list(window_hist_on->colors, (unsigned char *)"HISTMARK"))) {
      col_print(ln->right);
    } else {
      col_print(TIMELINEON);
    }
    pos = WinSize.start + window_hist_pos + num - cursor.row;
    
    line = window_hist_on->get_hist_end() - pos;
    if(line < 0)
      line += window_hist_on->get_hist_size();

    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    if(orgpos.col > len) {
      orgpos.col = len;
      move(&orgpos, FALSE);
    }
    window_hist_on->set_hist_mark(pos, cursor.col);
    
    buf[0] = *(window_hist_on->get_scrhistory(line) + cursor.col - 1);
    buf[1] = '\0';
#ifndef X_DISPLAY_MISSING
    if(MainDisplay)
      xprint(buf);
    else
#endif
      printf("%s", buf);
    
    check_color();
    move(&orgpos, TRUE);
  }
}

void window::hist_copy_clip()
{
  unsigned char buffer[BUFFER_SIZE], *str;
  int cursorpos, lines, num, line, len, amount, i, strpos;
  int startpos, startcol, endpos, endcol, tmprow;
  const listnode *ln;
  position orgpos;
  
  if(window_hist_on) {
    num = window_hist_on->get_hist_num();
    
    lines = WinSize.end - WinSize.start;
    if(get_time_on())
      lines--;
    
    if(num > lines)
      num = lines;
    
    cursorpos = WinSize.start + window_hist_pos + num - cursor.row;
    
    line = window_hist_on->get_hist_end() - cursorpos;
    if(line < 0)
      line += window_hist_on->get_hist_size();
    
    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    if(cursor.col > len) {
      cursor.col = len;
      move(&cursor, FALSE);
    }

    if(cursorpos == window_hist_on->get_hist_markp()) {
      if(window_hist_on->get_hist_markc() < cursor.col) {
	startcol = window_hist_on->get_hist_markc();
	endcol   = cursor.col;
      } else {
	startcol = cursor.col;
	endcol   = window_hist_on->get_hist_markc();
      }
      amount = endcol - startcol + 2;
      
      if(!(str = new unsigned char[amount])) {
	print_win_hist_line(FALSE, (unsigned char *)"OUT OF MEMORY");
	return;
      }
      strpos = 0;
      for(i = startcol - 1; i < endcol; i++)
	*(str + strpos++) = window_hist_on->get_scrhistory(line)[i];
      *(str + strpos) = '\0';

      if(selection)
	delete[] selection;
      selection = str;
      selection_len = strpos;

#ifndef X_DISPLAY_MISSING
      if(MainDisplay) {
	if(!XtOwnSelection(base->widget, XA_PRIMARY, CurrentTime,
			   convertSelectionProc, loseSelectionProc,
			   doneSelectionProc)) {
	  delete[] str;
	  selection = 0;
	  selection_len = 0;
	  return;
	}
      }
#endif
      
      orgpos = cursor;
      if(cursor.col != startcol)
	move(cursor.row, startcol, FALSE);
      
      if((ln = searchnode_list(window_hist_on->colors, (unsigned char *)"HISTMARK"))) {
	col_print(ln->right);
      } else {
	col_print(TIMELINEON);
      }
#ifndef X_DISPLAY_MISSING
      if(MainDisplay)
	xprint(selection);
      else
#endif
	printf("%s", selection);
      move(&orgpos, TRUE);
      check_color();
    } else {
      if(cursorpos < window_hist_on->get_hist_markp()) {
	startpos = window_hist_on->get_hist_markp();
	startcol = window_hist_on->get_hist_markc();
	endpos = cursorpos;
	endcol = cursor.col;
      } else {
	startpos = cursorpos;
	startcol = cursor.col;
	endpos = window_hist_on->get_hist_markp();
	endcol = window_hist_on->get_hist_markc();
      }
      
      line = window_hist_on->get_hist_end() - startpos;
      if(line < 0)
	line += window_hist_on->get_hist_size();
      amount = strlen((const char *)window_hist_on->get_scrhistory(line)) - startcol + 1;
      
      for(i = startpos - 1;
	  i > endpos;
	  i--) {
	line = window_hist_on->get_hist_end() - i;
	if(line < 0)
	  line += window_hist_on->get_hist_size();
	
	amount += strlen((const char *)window_hist_on->get_scrhistory(line));
      }
      amount += endcol + 1;
      
      if(amount >= BUFFER_SIZE) {
	print_win_hist_line(FALSE, (unsigned char *)"Too Long String To Clip");
	return;
      }
      
      if(!(str = new unsigned char[amount])) {
	print_win_hist_line(FALSE, (unsigned char *)"OUT OF MEMORY");
	return;
      }
      
      strpos = 0;
      
      orgpos = cursor;
      
      if((ln = searchnode_list(window_hist_on->colors, (unsigned char *)"HISTMARK"))) {
        col_print(ln->right);
      } else {
        col_print(TIMELINEON);
      }
      
      line = window_hist_on->get_hist_end() - startpos;
      if(line < 0)
        line += window_hist_on->get_hist_size();
      
      strpos = str_cat(str, strpos, amount,
		       window_hist_on->get_scrhistory(line) + startcol - 1);
      
      if(startpos > window_hist_pos && startpos <= window_hist_pos + num) {
	tmprow = WinSize.start + window_hist_pos + num - startpos;
	move(tmprow, startcol, FALSE);
#ifndef X_DISPLAY_MISSING
	if(MainDisplay)
	  xprint(window_hist_on->get_scrhistory(line) + startcol - 1);
	else
#endif
	  printf("%s", window_hist_on->get_scrhistory(line) + startcol - 1);
      }
      
      for(i = startpos - 1;
	  i > endpos;
	  i--) {
	line = window_hist_on->get_hist_end() - i;
	if(line < 0)
	  line += window_hist_on->get_hist_size();
	
	strpos = str_cat(str, strpos, amount,
			 window_hist_on->get_scrhistory(line));
	
	if(i > window_hist_pos && i <= window_hist_pos + num) {
	  move(WinSize.start + window_hist_pos + num - i, 1, FALSE);
#ifndef X_DISPLAY_MISSING
	  if(MainDisplay)
	    xprint(window_hist_on->get_scrhistory(line));
	  else
#endif
	    printf("%s", window_hist_on->get_scrhistory(line));
	}
      }
      
      line = window_hist_on->get_hist_end() - endpos;
      if(line < 0)
	line += window_hist_on->get_hist_size();
      
      len = strlen((const char *)window_hist_on->get_scrhistory(line));
      
      for(i = 0; i < endcol; i++) {
	*(str + strpos++) = window_hist_on->get_scrhistory(line)[i];
	*(buffer + i) = window_hist_on->get_scrhistory(line)[i];
      }
      *(str + strpos) = '\0';
      *(buffer + i) = '\0';

      if(selection)
	delete[] selection;
      selection = str;
      selection_len = strpos;
      
#ifndef X_DISPLAY_MISSING
      if(MainDisplay) {
	if(!XtOwnSelection(base->widget, XA_PRIMARY, CurrentTime,
			   convertSelectionProc, loseSelectionProc,
			   doneSelectionProc)) {
	  delete[] str;
	  selection = 0;
	  selection_len = 0;
	  return;
	}
      }
#endif

      if(endpos > window_hist_pos && endpos <= window_hist_pos + num) {
	move(WinSize.start + window_hist_pos + num - endpos, 1, FALSE);
#ifndef X_DISPLAY_MISSING
	if(MainDisplay)
	  xprint(buffer);
	else
#endif
	  printf("%s", buffer);
      }
      
      move(&orgpos, TRUE);
      check_color();
    }
    fflush(stdout);
    window_hist_clip = TRUE;
  }
}

void window::hist_rem_clip_color()
{
  unsigned char buffer[BUFFER_SIZE], buf2[2];
  int cursorpos, lines, num, line, len, i, strpos;
  int startpos, startcol, endpos, endcol, tmprow;
  position orgpos;
  const listnode *ln;
  
  if(window_hist_on && window_hist_clip) {
    window_hist_clip = FALSE;

    num = window_hist_on->get_hist_num();
    
    lines = WinSize.end - WinSize.start;
    if(get_time_on())
      lines--;
    
    if(num > lines)
      num = lines;
    
    cursorpos = WinSize.start + window_hist_pos + num - cursor.row;
    
    line = window_hist_on->get_hist_end() - cursorpos;
    if(line < 0)
      line += window_hist_on->get_hist_size();
    
    len = strlen((const char *)window_hist_on->get_scrhistory(line));
    if(cursor.col > len) {
      cursor.col = len;
      move(&cursor, FALSE);
    }
    
    check_color();
    
    orgpos = cursor;
    
    if(cursorpos == window_hist_on->get_hist_markp()) {
      if(window_hist_on->get_hist_markc() < cursor.col) {
	startcol = window_hist_on->get_hist_markc();
	endcol   = cursor.col;
      } else {
	startcol = cursor.col;
	endcol   = window_hist_on->get_hist_markc();
      }
      strpos = 0;
      for(i = startcol - 1; i < endcol; i++)
	*(buffer + strpos++) = window_hist_on->get_scrhistory(line)[i];
      *(buffer + strpos) = '\0';
      
      move(cursor.row, startcol, FALSE);
#ifndef X_DISPLAY_MISSING
      if(MainDisplay)
	xprint(buffer);
      else
#endif
	printf("%s", buffer);

    } else {
      if(cursorpos < window_hist_on->get_hist_markp()) {
	startpos = window_hist_on->get_hist_markp();
	startcol = window_hist_on->get_hist_markc();
	endpos = cursorpos;
	endcol = cursor.col;
      } else {
	startpos = cursorpos;
	startcol = cursor.col;
	endpos = window_hist_on->get_hist_markp();
	endcol = window_hist_on->get_hist_markc();
      }
      strpos = 0;
      
      if(startpos > window_hist_pos && startpos <= window_hist_pos + num) {
	line = window_hist_on->get_hist_end() - startpos;
	if(line < 0)
	  line += window_hist_on->get_hist_size();
	
	strpos = str_cat(buffer, strpos, sizeof(buffer),
			 window_hist_on->get_scrhistory(line) + startcol - 1);
	
	tmprow = WinSize.start + window_hist_pos + num - startpos;
	move(tmprow, startcol, FALSE);
#ifndef X_DISPLAY_MISSING
	if(MainDisplay)
	  xprint(window_hist_on->get_scrhistory(line) + startcol - 1);
	else
#endif
	  printf("%s", window_hist_on->get_scrhistory(line) + startcol - 1);
      }
      
      for(i = startpos - 1;
	  i > endpos;
	  i--) {
	if(i > window_hist_pos && i <= window_hist_pos + num) {
	  line = window_hist_on->get_hist_end() - i;
	  if(line < 0)
	    line += window_hist_on->get_hist_size();
	  
	  move(WinSize.start + window_hist_pos + num - i, 1, FALSE);
#ifndef X_DISPLAY_MISSING
	  if(MainDisplay)
	    xprint(window_hist_on->get_scrhistory(line));
	  else
#endif
	    printf("%s", window_hist_on->get_scrhistory(line));
	}
      }
      
      if(endpos > window_hist_pos && endpos <= window_hist_pos + num) {
	line = window_hist_on->get_hist_end() - endpos;
	if(line < 0)
	  line += window_hist_on->get_hist_size();
	
	len = strlen((const char *)window_hist_on->get_scrhistory(line));
	
	for(i = 0; i < endcol; i++) {
	  *(buffer + i) = window_hist_on->get_scrhistory(line)[i];
	}
	*(buffer + i) = '\0';
	
	move(WinSize.start + window_hist_pos + num - endpos, 1, FALSE);
#ifndef X_DISPLAY_MISSING
	if(MainDisplay)
	  xprint(buffer);
	else
#endif
	  printf("%s", buffer);
      }
    }
    
    if(window_hist_on->get_hist_markp() > window_hist_pos &&
       window_hist_on->get_hist_markp() <= window_hist_pos + num) {
      
      move(WinSize.start + window_hist_pos + num -
	   window_hist_on->get_hist_markp(),
	   window_hist_on->get_hist_markc(), FALSE);
      
      if((ln = searchnode_list(window_hist_on->colors, (unsigned char *)"HISTMARK"))) {
        col_print(ln->right);
      } else {
        col_print(TIMELINEON);
      }
      
      line = window_hist_on->get_hist_end() - window_hist_on->get_hist_markp();
      if(line < 0)
	line += window_hist_on->get_hist_size();
      
      buf2[0] = *(window_hist_on->get_scrhistory(line) +
		  window_hist_on->get_hist_markc() - 1);
      buf2[1] = '\0';
#ifndef X_DISPLAY_MISSING
      if(MainDisplay)
	xprint(buf2);
      else
#endif
	printf("%s", buf2);
      
      check_color();
    }
    move(&orgpos, TRUE);
    
    fflush(stdout);
  }
}

int window::start_win_history(session *ses)
{
  int WinEnd;
  
  if(!window_hist_on) {
    WinEnd = WinSize.end;
    if(get_time_on())
      WinEnd--;

    if(WinEnd > WinSize.start) {
      remove_user_text();
      *INPUTBUF = '\0';
      INPUTPOS = 0;
      CURSOR = 0;
      print_prompt(FALSE);
      
      window_hist_on = ses;
      redraw();
      return(TRUE);
    } else {
      ses->print((const uchar *)"#Too small window for history.\n\r");
    }
  }
  return(FALSE);
}

void window::exit_win_history()
{
  if(window_hist_on) {
    window_hist_pos = 0;
    window_hist_on  = 0;
    
    *INPUTBUF = '\0';
    INPUTPOS = 0;
    CURSOR   = 0;
    redraw();
  }
}

void window::max_size()
{
  int tmp;
  session *ses, *nextses;
  
  tmp = WinSize.end - Scroll.end;
  if(next) {
    WinSize.end = next->WinSize.start - 1;
  } else {
    WinSize.end = get_rows();
  }
  Scroll.end  = WinSize.end - tmp;
  redraw();
  for(ses = session::get_sessions(); ses; ses = nextses) {
    nextses = ses->get_next();
    if(ses->get_output() == this)
      ses->send_naws();
  }
}

int window::change_size(const int newsize)
{
  int oldsize = WinSize.end - WinSize.start + 1;
  int sizediff, newlines = 0, lines;
  window *prewin = 0;
  session *ses, *nextses;

  assert(newsize >= 3);
  
  if(newsize != oldsize) {
    clear();

    sizediff = WinSize.end - Scroll.end;
    
    if(newsize > oldsize) {               // bigger window
      newlines = newsize - oldsize;
      
      if(next) {
	lines = next->get_size()->end - next->get_size()->start + 1;
	if(lines - 3 > 0) {
	  if(newlines >= lines - 3) {
	    WinSize.end = next->get_size()->end - 3;
	    newlines -= lines - 3;
	  } else {
	    WinSize.end += newlines;
	    newlines = 0;
	  }
	  Scroll.end  = WinSize.end - sizediff;
	  next->WinSize.start = WinSize.end + 1;
	  next->Scroll.start  = next->WinSize.start;
	  next->redraw();
	  
	  for(ses = session::get_sessions(); ses; ses = nextses) {
	    nextses = ses->get_next();
	    if(ses->get_output() == next)
	      ses->send_naws();
	  }
	}
      } else {
	lines = get_rows() - WinSize.end;
	if(lines > 0) {
	  if(newlines >= lines) {
	    WinSize.end = get_rows();
	    newlines -= lines;
	  } else {
	    WinSize.end += newlines;
	    newlines = 0;
	  }
	  Scroll.end  = WinSize.end - sizediff;
	}
      }
      if(newlines > 0) {
	if(base->get_windows() != this) {
	  for(prewin = base->get_windows();
	      prewin->next && prewin->next != this;
	      prewin = prewin->get_next())
	    ;
	}
	if(prewin) {
	  lines = prewin->get_size()->end - prewin->get_size()->start + 1;

	  if(lines - 3 > 0) {
	    sizediff = prewin->WinSize.end - prewin->Scroll.end;
	    
	    if(newlines >= lines - 3) {
	      WinSize.start = prewin->WinSize.start + 3;
	      newlines -= lines - 3;
	    } else {
	      WinSize.start -= newlines;
	      newlines = 0;
	    }
	    Scroll.start  = WinSize.start;
	    prewin->WinSize.end = WinSize.start - 1;
	    prewin->Scroll.end  = prewin->WinSize.end - sizediff;
	    prewin->redraw();
	    for(ses = session::get_sessions(); ses; ses = nextses) {
	      nextses = ses->get_next();
	      if(ses->get_output() == prewin)
		ses->send_naws();
	    }
 	  }
	}
      }
    } else {                             // smaller window
      if(next) {
	WinSize.end = WinSize.start + newsize - 1;
	Scroll.end  = WinSize.end - sizediff;
	next->WinSize.start = WinSize.end + 1;
	next->Scroll.start  = next->WinSize.start;
	next->redraw();
	for(ses = session::get_sessions(); ses; ses = nextses) {
	  nextses = ses->get_next();
	  if(ses->get_output() == next)
	    ses->send_naws();
	}
      } else if(this != base->get_windows()) {
	for(prewin = base->get_windows();
	    prewin->next && prewin->next != this;
	    prewin = prewin->get_next())
	  ;
	
	assert(prewin);
	
	WinSize.start = WinSize.end - newsize + 1;
	Scroll.start  = WinSize.start;
	
	cursor.col = 1;
	cursor.row = WinSize.start;
	
	sizediff = prewin->WinSize.end - prewin->Scroll.end;
	
	prewin->WinSize.end = WinSize.start - 1;
	prewin->Scroll.end  = prewin->WinSize.end - sizediff;
	prewin->redraw();
	for(ses = session::get_sessions(); ses; ses = nextses) {
	  nextses = ses->get_next();
	  if(ses->get_output() == prewin)
	    ses->send_naws();
	}
      }
    }
    redraw();
    for(ses = session::get_sessions(); ses; ses = nextses) {
      nextses = ses->get_next();
      if(ses->get_output() == this)
	ses->send_naws();
    }
  }
  return(newlines);
}

void window::set_active_ses(session *ses)
{
  if(window::active == this &&
     !ses->is_set_flag(SES_FLAG_LINEMODE) &&
     ses->get_socket() >= 0) {
    remove_user_text();
    *INPUTBUF = '\0';
    INPUTPOS  = 0;
    CURSOR    = 0;
    master_table.utexrem = FALSE;
  }
  active_ses = ses;
}

void window::next_window_active(window *newact, int focusin)
{
  window *win = 0;
  basewindow *oldbase;
  session *newactive;
#ifndef X_DISPLAY_MISSING
  XWindowAttributes winattr;
#endif
  
  if(!newact) {
    if(next)
      win = next;
    else {
      if(base->get_next())
	win = base->get_next()->get_windows();
      else
	win = basewindow::get_bases()->get_windows();
    }
  } else
    win = newact;
  
  if(window::active != win) {
    oldbase = window::active->base;
    
    window::active->remove_user_text();
    
    if(!win->active_ses) {
      newactive = master_table.main_session;
      if(newactive->get_input()->get_active_ses() == newactive)
	newactive->get_input()->check_session();
      
      win->set_active_ses(newactive);
      newactive->set_input(win);
    } else {
      if(!win->active_ses->is_set_flag(SES_FLAG_LINEMODE) &&
	 win->active_ses->get_socket() >= 0) {
	*INPUTBUF = '\0';
	INPUTPOS  = 0;
	CURSOR    = 0;
	master_table.utexrem = FALSE;
      }
    }
    
    window::active->print_time(TRUE);
    win->print_time(TRUE);
    
    if(window::active->window_hist_on)
      window::active->print_win_hist_line(FALSE);
    
    window::active = win;
    
    if(win->window_hist_on) {
      win->print_win_hist_line(TRUE);
    } else {
      win->move(win->get_cursor(), FALSE);
      win->print_prompt(FALSE, FALSE);
    }
    
#ifndef X_DISPLAY_MISSING
    if(MainDisplay) {
      if(oldbase != win->base && focusin) {
	if(XGetWindowAttributes(MainDisplay,
				XtWindow(win->base->widget),
				&winattr)) {
	  if(winattr.map_state == IsViewable) {
	    XRaiseWindow(MainDisplay, XtWindow(win->base->widget));
	    XSetInputFocus(MainDisplay, XtWindow(win->base->widget),
			   RevertToPointerRoot, CurrentTime);
	  } else {
	    XMapRaised(MainDisplay, XtWindow(win->base->widget));
	  }
	}
      }
      win->xprint_cursor(TRUE);
    }
#endif
  }
}

// **************************************
//
// **************************************

void window::check_session(const session *o_ses)
{
  session *ses = 0;
  
  if(o_ses) {
    for(ses = session::get_sessions(); ses; ses = ses->get_next())
      if(ses != o_ses && ses->get_input() == this)
	break;
  } else {
    for(ses = session::get_sessions(); ses; ses = ses->get_next())
      if(ses->get_input() == this)
	break;
  }    
  
  active_ses = ses;
}

// *****************************
//  function to get screen size 
// *****************************

void get_screen_size(int &ScrCols, int &ScrRows)
{
  unsigned char buf[1024], *bufpos;
  int col, row;
  winsize WSize;

#ifndef X_DISPLAY_MISSING
  assert(!MainDisplay);
#endif
  
  ioctl(0, TIOCGWINSZ, &WSize);
  
  ScrRows = WSize.ws_row;
  ScrCols = WSize.ws_col;
  
  if(ScrCols < 40 || ScrRows < 2) {
    nocbreak_echo();
    printf("%s", AUTOWARPON);
    printf("Screen Size: %dx%d\n\r", WSize.ws_col , WSize.ws_row);
    printf("Couldn't get screen size or too small screen!!!\n\r"
	   "Smallest allowed screen size is: 40x2\n\r");
    while(ScrCols < 40 || ScrRows < 2) {
      printf("Type screen size (default: %dx%d): ", 
	     default_chars_per_line, default_lines_in_scr);
      for(bufpos = buf; (*bufpos = getchar()) != '\n'; bufpos++)
	;
      *bufpos = '\0';
      if(sscanf((const char *)buf, " %d x %d", &col, &row) != 2) {
	printf("\n\rUsing defaults values.\n\r");
	ScrCols = default_chars_per_line;
	ScrRows = default_lines_in_scr;
      } else {
	if(col >= 40 && row >= 2) {
	  ScrCols = col;
	  ScrRows = row;
	  default_chars_per_line = col;
	  default_lines_in_scr = row;
	}
      }
    }
    cbreak_noecho();
  }
  printf("%s", AUTOWARPOFF);
}

/* ***************************************************
 *     function to check screen size                 *
 *************************************************** */

void get_screen_size(void)
{
  int oldrows, oldcols;
  
#ifndef X_DISPLAY_MISSING
  assert(!MainDisplay);
#endif
  
  oldrows = basewindow::get_bases()->rows;
  oldcols = basewindow::get_bases()->cols;
  window::max_scroll_region();
  get_screen_size(basewindow::get_bases()->cols,
		  basewindow::get_bases()->rows);
  if(basewindow::get_bases()->rows != oldrows) {
    basewindow::get_bases()->check_win_sizes();
  }
  redraw_screen();
}

/* ********************************************************
 *   clear screen and print text again from scr history   *
 ******************************************************** */

void redraw_screen(void)
{
  window *tmp;
  
#ifndef X_DISPLAY_MISSING
  if(MainDisplay) {
    window::get_active_win()->redraw();
  } else {
#endif
    printf("%s", CLEARSCREEN);
    fflush(stdout);
    for(tmp = basewindow::get_bases()->get_windows();
	tmp;
	tmp = tmp->get_next())
      tmp->redraw();
    
    tmp = window::get_active_win();
    tmp->scroll_region(tmp->get_scroll());
#ifndef X_DISPLAY_MISSING
  }
#endif
}

// **********************
//   print all prompts
// **********************

void prompt(void)
{
  window *win;
  basewindow *base;
  
  for(base = basewindow::get_bases(); base; base = base->get_next())
    for(win = base->get_windows(); win; win = win->get_next())
      win->print_prompt(TRUE);
}

/* **************************************
 *  cbreak and echo                     *
 ************************************** */

#ifndef _POSIX_VDISABLE
# ifdef HAVE_SYS_PARAM_H
#  include <sys/param.h>        /* pick up VDISABLE definition, mayby */
# endif
# ifdef VDISABLE
#  define _POSIX_VDISABLE VDISABLE
# else
#  define _POSIX_VDISABLE ((char)'\377')
# endif
#endif

void cbreak_noecho(void)
{
#ifndef X_DISPLAY_MISSING
  assert(!MainDisplay);
#endif

#ifdef HAVE_TERMIOS_H
  struct termios cbreak_state;
  
  cbreak_state = original_state;
  cbreak_state.c_iflag &= ~(IXON | IXOFF);
  cbreak_state.c_lflag &= ~(ICANON | ISIG | ECHO);
  cbreak_state.c_cc[VMIN] = 1;
  cbreak_state.c_cc[VTIME] = 0;
  tcsetattr(0, TCSANOW, &cbreak_state);
#else
  struct sgttyb cbreak_state;
  struct tchars cbreak_chars;
  
  cbreak_state = original_state;
  cbreak_state.sg_flags |= CBREAK;
  cbreak_state.sg_flags |= RAW;
  cbreak_state.sg_flags &= ~ECHO;
  ioctl(0, TIOCSETP, &cbreak_state);
  cbreak_chars = original_chars;
  cbreak_chars.t_startc = _POSIX_VDISABLE;
  cbreak_chars.t_stopc  = _POSIX_VDISABLE;
  ioctl(0, TIOCSETC, &cbreak_chars);
#endif
}

void nocbreak_echo(void)
{
#ifndef X_DISPLAY_MISSING
  assert(!MainDisplay);
#endif

#ifdef HAVE_TERMIOS_H
  tcsetattr(0, TCSANOW, &original_state);
#else
  ioctl(0, TIOCSETP, &original_state);
  ioctl(0, TIOCSETC, &original_chars);
#endif
}

// *********
//  termcap
// *********

const uchar *check_tertmcap_str(const char *str)
{
  const char *tmpstr;
  char *hold_pos;
  int i;

  if(!str)
    return(0);
  
  while(isdigit(*str) || *str == '*' || *str == '.')
    // remove '.', numbers and '*' before control string
    str++;
  
  tmpstr = str;
  
  while(*tmpstr) {
    // remove curses terminfo padding characters
    
    if(*tmpstr == '$' && *(tmpstr + 1) == '<') {   
      hold_pos = (char *)tmpstr;
      tmpstr += 2;
      
      while(isdigit(*tmpstr) || *tmpstr == '*' || *tmpstr == '.')
	tmpstr++;
      
      if(*tmpstr == '>') {
	tmpstr++;
	
	if(*tmpstr) {
	  if(hold_pos == str) {
	    str = tmpstr;
	  } else {
	    for(i = 0; *(tmpstr + i); i++)
	      *(hold_pos + i) = *(tmpstr + i);
	    *(hold_pos + i) = '\0';
	    tmpstr = hold_pos;
	  }
	} else {
	  *hold_pos = '\0';
	}
      } else
	tmpstr++;
    } else
      tmpstr++;
  }
  
  if(!*str)
    return(0);
  
  return((const uchar *)str);
}

void init_screen(const int no_xw)
{
#ifndef X_DISPLAY_MISSING
  char *oargv[3] = {
    "HUT TINTIN",
    "",
    ""
    };
  int oargc = 1;
  
  XtToolkitInitialize();
  MainContext = XtCreateApplicationContext();
  
  if(!no_xw) 
    MainDisplay = XtOpenDisplay(MainContext, NULL, NULL, "HUT TINTIN",
				NULL, 0, &oargc, oargv);
  else
    MainDisplay = 0;
  
  if(MainDisplay) {
    environment_list.add((const uchar *)"TERM", DEF_TERM_TYPE);
    GetColors();
  } else {
#endif
#ifdef HAVE_TERMIOS_H
    tcgetattr(0, &original_state);
#else
    ioctl(0, TIOCGETP, &original_state);
    ioctl(0, TIOCGETC, &original_chars);
#endif
    cbreak_noecho();
#ifndef X_DISPLAY_MISSING
  }
#endif
}

void init_termcap()
{
#ifdef HAVE_LIBTERMCAP
  char errmsg[256];
  static char tbuf[1024];
  static char cbuf[1024];
  char *_tptr, *term;
  
  int len;
  
#ifndef X_DISPLAY_MISSING
  if(!MainDisplay) {
#endif

    if((term = getenv("TERM"))) {
      len = tgetent(cbuf, term);
      
      if(len > 0) {
	_tptr = tbuf;
	
	CEL         = check_tertmcap_str(tgetstr("ce", &_tptr));
	CES         = check_tertmcap_str(tgetstr("cd", &_tptr));
	CLEARSCREEN = check_tertmcap_str(tgetstr("cl", &_tptr));
	HOME        = check_tertmcap_str(tgetstr("ho", &_tptr));
	CUP         = check_tertmcap_str(tgetstr("up", &_tptr));
	CDOWN       = check_tertmcap_str(tgetstr("do", &_tptr));
	FORWARD     = check_tertmcap_str(tgetstr("nd", &_tptr));
	BACKWARD    = check_tertmcap_str(tgetstr("le", &_tptr));
	RESETCOLOR  = check_tertmcap_str(tgetstr("me", &_tptr));
	INSERTLINE  = check_tertmcap_str(tgetstr("al", &_tptr));
	VBELL       = check_tertmcap_str(tgetstr("vb", &_tptr));
	MOVE        = check_tertmcap_str(tgetstr("cm", &_tptr));
	
	TABSIZE     = tgetnum("it");
	
	add_key_string(check_tertmcap_str(tgetstr("kh",&_tptr)),func_home);
	add_key_string(check_tertmcap_str(tgetstr("kI",&_tptr)),
		       func_insert_clip);
	add_key_string(check_tertmcap_str(tgetstr("kH",&_tptr)),func_end);
	add_key_string(check_tertmcap_str(tgetstr("kP",&_tptr)),func_pageup);
	add_key_string(check_tertmcap_str(tgetstr("kN",&_tptr)),func_pagedown);
	add_key_string(check_tertmcap_str(tgetstr("ku",&_tptr)),func_arrowup);
	add_key_string(check_tertmcap_str(tgetstr("kd",&_tptr)),func_arrowdown);
	add_key_string(check_tertmcap_str(tgetstr("kr",&_tptr)),func_arrowright);
	add_key_string(check_tertmcap_str(tgetstr("kl",&_tptr)),func_arrowleft);
	add_key_string(check_tertmcap_str(tgetstr("k1",&_tptr)),func_mark);
	add_key_string(check_tertmcap_str(tgetstr("k2",&_tptr)),func_copy_clip);
	
	
      } else {
	if(len == 0) {
	  sprintf(errmsg, "No termcap entry for %s", term);
	  filelog(errmsg);
	} else {
	  filelog("No /etc/termcap present!");
	}
      }
    } else {
      filelog("Environment variable TERM not set");
    }
#ifndef X_DISPLAY_MISSING
  }
#endif
#endif /* ifdef HAVE_LIBTERMCAP */
  
  /* use vt100/ansi codes if didn't got from termcap */
  
  if(!CEL) CEL = (const uchar *)"\033[0K";    /* clear end of line         */
  if(!CSL) CSL = (const uchar *)"\033[1K";    /* clear start of line       */
  if(!CL)  CL  = (const uchar *)"\033[2K";    /* clear line                */
  if(!CES) CES = (const uchar *)"\033[0J";    /* clear end of screen       */
  if(!CSS) CSS = (const uchar *)"\033[1J";    /* clear start of screen     */
  if(!CLEARSCREEN) CLEARSCREEN = (const uchar *)"\033[H\033[2J"; /* clear screen              */
  if(!AUTOWARPON)  AUTOWARPON  = (const uchar *)"\033[?7h"; /* turns autowarp on         */
  if(!AUTOWARPOFF) AUTOWARPOFF = (const uchar *)"\033[?7l"; /* turns autowarp off        */
  if(!HOME)      HOME      = (const uchar *)"\033[H";
  if(!CUP)       CUP       = (const uchar *)"\033[A";
  if(!CDOWN)     CDOWN     = (const uchar *)"\033[B";
  if(!FORWARD)   FORWARD   = (const uchar *)"\033[C";
  if(!BACKWARD)  BACKWARD  = (const uchar *)"\033[D";
  if(!MAXSCROLLSIZE) MAXSCROLLSIZE = (const uchar *)"\033[r";
  if(!RESETCOLOR) RESETCOLOR = (const uchar *)"\033[0m";
  if(!INSERTLINE) INSERTLINE = (const uchar *)"\033[L";
  
  if(TABSIZE <= 0)  TABSIZE  = 8;
}

// ************************
//   command split window
// ************************

session *comm_split(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], name[BUFFER_SIZE];
  int value = 0, lines;
  window *win, *win2;
  const session *sesptr;

  arg = get_arg_stop_spaces(arg, name);
  arg = get_arg_with_spaces(arg, buf);
  if(*buf) {
    value = string_to_number(buf, ses);
    
    win = ses->get_input();

    lines = win->get_size()->end - win->get_size()->start + 1;
    
    if(value < 3) {
      ses->print((const uchar *)"#You cannot make that small windows.\n\r");
    } else if(value > lines - 3) {
      ses->print((const uchar *)"#Too BIG new window\n\r");
    } else {
      win2 = find_window(name);
      if(win2) {
	ses->print((const uchar *)"#Already that named window\n\r");
	return(ses);
      }
      for(sesptr = master_table.main_session;
	  sesptr;
	  sesptr = sesptr->get_next())
	if(strcmp((const char *)sesptr->get_name(), (const char *)name) == 0) {
	  ses->print((const uchar *)"#Already that named session.\n\r");
	  return(ses);
	}
      
      ses->print((const uchar *)"#Ok.\n\r");
      ses->print_prompt(TRUE);
      win2 = new window(name, value, win);
      ses = window::get_active_win()->get_active_ses();
    }
  } else {
    ses->print(help_split);
  }
  return(ses);
}

// **********************
//   command open window
// **********************

session *comm_open(const unsigned char *arg, session *ses)
{
#ifndef X_DISPLAY_MISSING
  const char *geometry = 0, *fontname = 0;
  uchar name[BUFFER_SIZE], geobuf[BUFFER_SIZE], fontbuf[BUFFER_SIZE];
  window *win;
  basewindow *base;
  const session *sesptr;
  
  if(!MainDisplay) {
    ses->print((const uchar *)"#Couldn't open Display.\n\r");
    ses->set_status(FALSE);
    return(ses);
  }
  
  arg = get_arg_stop_spaces(arg, name);
  while(*name == '-') {
    switch(*(name + 1)) {
    case 'f':
    case 'F':
      get_arg_stop_spaces(name + 2, fontbuf);
      if(!*fontbuf)
	arg = get_arg_stop_spaces(arg, fontbuf);
      if(*fontbuf)
	fontname = (const char *)fontbuf;
      break;
    case 'g':
    case 'G':
      get_arg_stop_spaces(name + 2, geobuf);
      if(!*geobuf)
	arg = get_arg_stop_spaces(arg, geobuf);
      if(*geobuf)
	geometry = (const char *)geobuf;
      break;
    default:
      ses->print((const uchar *)"#Unknow argument: '");
      ses->print(name);
      ses->print((const uchar *)"'\n\r");
      break;
    }
    arg = get_arg_stop_spaces(arg, name);
  }
  
  if(*name) {
    win = find_window(name);
    if(win) {
      ses->print((const uchar *)"#Already that named window\n\r");
      ses->set_status(FALSE);
      return(ses);
    }
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      if(strcmp((const char *)sesptr->get_name(), (const char *)name) == 0) {
	ses->print((const uchar *)"#Already that named session.\n\r");
	ses->set_status(FALSE);
	return(ses);
      }
    
    ses->print((const uchar *)"#Ok.\n\r");
    base = new basewindow(name, geometry, fontname, ses);
    if(!base->widget) { // didn't get window
      delete base;
      ses->set_status(FALSE);
      return(ses);
    }
    ses->set_status(TRUE);
  } else {
    ses->print(help_open);
    ses->set_status(FALSE);
  }
#else
  ses->print((const uchar *)"#HT not compiled with xwindow code.\n\r");
  ses->set_status(FALSE);
#endif

  return(ses);
}

// ***************************
//   command redirect output
// ***************************


session *comm_output(const unsigned char *arg, session *ses)
{
  unsigned char name[BUFFER_SIZE];
  window *win;
  
  arg = get_arg_with_spaces(arg, name);
  if(*name) {
    win = find_window(name);
    if(win && win != NULL_WIN) {
      ses->set_output(win);
      ses->print((const uchar *)"#Ok.\n\r");
      ses = ses->send_naws();
    } else {
      ses->print((const uchar *)"#NO such named window\n\r");
    }
  } else {
    ses->print(help_output);
  }
  return(ses);
}

// **************************
//   command redirect input
// **************************

session *comm_input(const unsigned char *arg, session *ses)
{
  unsigned char name[BUFFER_SIZE];
  window *win;
  
  arg = get_arg_with_spaces(arg, name);
  if(*name) {
    win = find_window(name);
    if(win && win != NULL_WIN) {
      if(win != ses->get_input()) {
	if(ses == ses->get_input()->get_active_ses())
	  ses->get_input()->check_session(ses);
	
	ses->set_input(win);
	win->set_active_ses(ses);
	window::get_active_win()->next_window_active(win);
	if(!win->get_hist_on()) {
	  win->print((const uchar *)"\r");
	  win->set_ptexrem();
	}
      }
      ses->print((const uchar *)"#Ok.\n\r");
    } else {
      ses->print((const uchar *)"#NO such named window\n\r");
    }
  } else {
    ses->print(help_input);
  }
  return(ses);
}

// *************************
//   command close window
// *************************

session *comm_close(const unsigned char *arg, session *ses)
{
  unsigned char name[BUFFER_SIZE];
  window *win;
  
  arg = get_arg_with_spaces(arg, name);
  if(*name) {
    win = find_window(name);
    if(win && win != NULL_WIN) {
      delete win;
      redraw_screen();
    } else {
      ses->print((const uchar *)"#NO such named window\n\r");
    }
  } else {
    delete ses->get_input();
    redraw_screen();
  }
  return(ses);
}

// ******************************
//   command window information
// ******************************

session *comm_wininfo(const unsigned char *, session *ses)
{
  unsigned char buf[BUFFER_SIZE], color[BUFFER_SIZE];
  unsigned char buf2[BUFFER_SIZE];
  unsigned short tmp = default_color;
  window *win;
  basewindow *base;

  for(base = basewindow::get_bases(); base; base = base->get_next()) {
    for(win = base->get_windows(); win; win = win->get_next()) {
      bits_to_color(tmp, win->get_colormode(), color);
      if(*color)
	color_to_text(color, buf2, 0);
      else
	sprintf((char *)buf2, "reset");
      sprintf((char *)buf,
	      "#Window Name     : %s\n\r"
#ifndef X_DISPLAY_MISSING
	      "#  Size          : %2d%s lines (Screen lines: %2d-%2d) %2d chars/line\n\r"
#else
	      "#  Size          : %2d%s lines (Screen lines: %2d-%2d)\n\r"
#endif
	      "#  Scroll        : %2d      lines (Screen lines: %2d-%2d)\n\r"
	      "#  Cursor        : %dCol, %dRow\n\r"
	      "#  Colors:       : %s\n\r"
	      "#  Active session: %s\n\r",
	      win->get_name(),
	      (win->get_size()->end -
	       win->get_size()->start + 1 -
	       win->get_time_on()),
	      (win->get_time_on() ? " (+1)" : "     "),
	      win->get_size()->start,
	      win->get_size()->end,
#ifndef X_DISPLAY_MISSING
	      win->get_cols(),
#endif
	      (win->get_scroll()->end - win->get_scroll()->start + 1),
	      win->get_scroll()->start,
	      win->get_scroll()->end,
	      win->get_cursor()->col,
	      win->get_cursor()->row,
	      buf2,
	      (win->get_active_ses() ?
	       (char *)win->get_active_ses()->get_name() :
	       "(none)"));
      ses->print(buf);
    }
  }
  sprintf((char *)buf, "#Session Input/Output windows: %s/%s\n\r",
	  ses->get_input()->get_name(),
	  ses->get_output()->get_name());
  ses->print(buf);
  return(ses);
}

// ********************
// command winsize code
// ********************

session *comm_winsize(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];
  int maxsize, minsize = TRUE, newsize = 0, lines;
  window *pre, *win;
  
  arg = get_arg_with_spaces(arg, buffer);
  if(*buffer) {
    win = ses->get_output();
    maxsize = win->get_size()->end - win->get_size()->start + 1;
    if(win->get_next()) {
      lines = (win->get_next()->get_size()->end -
	       win->get_next()->get_size()->start + 1);
      maxsize += lines - 3;
      minsize = FALSE;
    } else {
      maxsize += win->get_rows() - win->get_size()->end;
    }
    if(win != win->get_base()->get_windows()) {
      for(pre = win->get_base()->get_windows();
	  pre->get_next() && pre->get_next() != win;
	  pre = pre->get_next())
	;
      if(pre) {
	lines = (pre->get_size()->end - pre->get_size()->start + 1);
	maxsize += lines - 3;
	minsize = FALSE;
      }
    }

    if(stricmp((const uchar *)"min", buffer) == 0)
      newsize = 3;
    else if(stricmp((const uchar *)"max", buffer) == 0)
      newsize = maxsize;
    else
      newsize = string_to_number(buffer, ses);
    
    if(minsize)
      ses->print((const uchar *)"#Only one window so you cannot change it's size.\n\r");
    else if(newsize < 3)
      ses->print((const uchar *)"#Too Small window size.\n\r");
    else if(newsize > maxsize)
      ses->print((const uchar *)"#Too Big window size.\n\r");
    else {
      ses->print((const uchar *)"#Ok.\n\r");
      ses->print_prompt(TRUE);
      if(win->change_size(newsize))
	ses->print((const uchar *)"#Hmmm... problems somewhere...:( (make bug report..)\n\r");
    }
  } else {
    ses->print(help_winsize);
  }
 
  return(ses);
}
  
// *******************************
//  code to find window with name
// *******************************

window *find_window(const unsigned char *name)
{
  window *win;
  basewindow *base;
  
  if(strcmp("null", (const char *)name) == 0)
    return(NULL_WIN);
  
  for(base = basewindow::get_bases(); base; base = base->get_next()) {
    for(win = base->get_windows(); win; win = win->get_next()) {
      if(strcmp((const char *)win->get_name(), (const char *)name) == 0)
	return(win);
    }
  }
  
  return(0);
}
