/*********************************************************************/
/* File: telnet_opt.cc - functions related telnet options            */
/*                            HUT TINTIN                             */
/*                     coded by Ossi Tulonen                         */
/*********************************************************************/
/*
 * Copyright (c) 1996 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "structs.h"

#include <arpa/inet.h>

#include <assert.h>

#include "session.h"

#include "environment.h"
#include "net.h"
#include "telnet_opt.h"
#include "utility.h"

int session::get_will_wont(int telopt) const
{
  if(telopt < 0 || telopt >= NTELOPTS)
    return(WONT);

  return(will_status[telopt]);
}

int session::get_do_dont(int telopt) const
{
  if(telopt < 0 || telopt >= NTELOPTS)
    return(DONT);
  
  return(do_status[telopt]);
}

int session::send_wont(int telopt)
{
  uchar wont_str[4] = {IAC, WONT, 0, '\0'};
  
  if(telopt < 0 || telopt >= NTELOPTS)
    return(FALSE);
  
  if(will_status[telopt] == WILL) {
    will_status[telopt] = WONT;
    wont_str[2] = telopt;
    
    if(write_socket(wont_str, get_socket(), 3) < 0) {
      print_errmsg("write socket");
      rem_net();
      return(-1);
    }
    return(TRUE);
  }
  return(FALSE);
}

// ****************************************************
//  code to parse & respont terminal type send askings
// ****************************************************

int parse_SB_TTYPE(uchar telqual, session *ses)
{
  uchar buf[256];
  const uchar *termtype;
  int pos;
  
  if(telqual != TELQUAL_SEND) /* don't do anything with these */
    return(FALSE);
  
  pos = 0;
  buf[pos++] = IAC;
  buf[pos++] = SB;
  buf[pos++] = TELOPT_TTYPE;
  buf[pos++] = TELQUAL_IS;
  
  if((termtype = environment_list.find_value((const uchar *)"TERM"))) {
    pos = str_cat(buf, pos, sizeof(buf), termtype);
  } else {
    pos = str_cat(buf, pos, sizeof(buf), DEF_TERM_TYPE);
  }
  
  buf[pos++] = IAC;
  buf[pos++] = SE;
  
  if(write_socket(buf, ses->get_socket(), pos) < 0) {
    ses->print_errmsg("write socket");
    ses->rem_net();
    return(-1);
  }
  return(TRUE);
}

// ************************************************
//  code to parse & respont X Display send askings
// ************************************************

int parse_SB_XDISPLOC(uchar telqual, session *ses)
{
  uchar buf[256];
  const uchar *dispname;
  int pos;
  
  if(telqual != TELQUAL_SEND) /* don't do anything with these */
    return(FALSE);
  
  pos = 0;
  buf[pos++] = IAC;
  buf[pos++] = SB;
  buf[pos++] = TELOPT_XDISPLOC;
  buf[pos++] = TELQUAL_IS;
  
  if((dispname = environment_list.find_value((const uchar *)"DISPLAY"))) {
    pos = str_cat(buf, pos, sizeof(buf), dispname);
    
  } else { /* we don't anymore have location */
    return(ses->send_wont(TELOPT_XDISPLOC));
  }
  
  buf[pos++] = IAC;
  buf[pos++] = SE;
  
  if(write_socket(buf, ses->get_socket(), pos) < 0) {
    ses->print_errmsg("write socket");
    ses->rem_net();
    return(-1);
  }
  return(TRUE);
}

// **********************************************
//  code to parse & respont environ send askings
// **********************************************

int parse_SB_ENVIRON(const uchar *str, const uchar *str_end, session *ses)
{
  uchar telopt, letter, name[BUFFER_SIZE];
  int var_nr, send_well_known = FALSE, send_user_var = FALSE;
  env_elem *ee;

  telopt = *str++;
  
  assert(telopt == TELOPT_NEW_ENVIRON || telopt == TELOPT_OLD_ENVIRON);
  
  if(str >= str_end || *str != TELQUAL_SEND)
    return(FALSE);
  str++;
  
  environment_list.start_sb_is(telopt, var_nr);
  
  if(str == str_end) {
    send_well_known = TRUE;
  }
  
  while(str < str_end) {
    letter = *str++;
    str = environment_list.parse_name(str, str_end, name);
    
    if(!*name) {
      if(letter == var_nr)
	send_well_known = TRUE;
      else if(letter == ENV_USERVAR)
	send_user_var = TRUE;
      
    } else {
      if((ee = environment_list.find(name))) {
	if((letter == var_nr && ee->is_well_known()) ||
	   (letter == ENV_USERVAR && !ee->is_well_known()))
	  environment_list.add_sb_is(ee);
	
      }
    }
  }

  if(send_well_known) {
    environment_list.add_all_sb_is(FALSE);
  }
  
  if(send_user_var) {
    environment_list.add_all_sb_is(TRUE);
  }

  return(environment_list.end_sb_is(TRUE, ses));
}
